"""VisDrone2019 object detection dataset."""
from __future__ import absolute_import
from __future__ import division
import os
import logging
import numpy as np
import mxnet as mx
from ..base import VisionDataset


class VisDroneDetection(VisionDataset):
    """VisDrone2019-DET Dataset.

    Parameters
    ----------
    root : str, default '~/mxnet/datasets/visdrone'
        Path to folder storing the dataset.
    splits : list of str, default ['train']
        Candidates can be: train, val.
    transform : callable, default None
        A function that takes data and label and transforms them. Refer to
        :doc:`./transforms` for examples.

        A transform function for object detection should take label into consideration,
        because any geometric modification will require label to be modified.
    """
    CLASSES = ('pedestrian', 'person', 'bicycle', 'car', 'van',
               'truck', 'tricycle', 'awning-tricycle', 'bus',
               'motor', 'others')

    def __init__(self, root=os.path.join('~', '.mxnet', 'datasets', 'visdrone'),
                 splits=('train',),
                 transform=None):
        super(VisDroneDetection, self).__init__(root)
        self._im_shapes = {}
        self._root = os.path.expanduser(root)
        self._transform = transform
        self._splits = splits
        self._anno_path = os.path.join('{}', 'annotations', '{}.txt')
        self._image_path = os.path.join('{}', 'images', '{}.jpg')
        self._images_dir = os.path.join('{}', 'images')
        self._items = self._load_items(self._splits)
        self.num_classes = len(self.CLASSES)
        self.index_map = dict(zip(self.CLASSES, range(self.num_class)))

    def __str__(self):
        detail = ','.join([str(s) for s in self._splits])
        return self.__class__.__name__ + '(' + detail + ')'

    @property
    def classes(self):
        """Category names."""
        return type(self).CLASSES

    def __len__(self):
        return len(self._items)

    def __getitem__(self, idx):
        img_id = self._items[idx]
        img_path = self._image_path.format(*img_id)
        img = mx.image.imread(img_path, 1)
        height, width, _ = img.shape
        label = self._load_label(idx, height, width)
        if self._transform is not None:
            return self._transform(img, label)
        return img, label.copy()

    def _load_items(self, splits):
        """Load individual image indices from splits."""
        ids = list()
        for name in splits:
            root = os.path.join(self._root, 'VisDrone2019-DET-' + name)
            images_dir = self._images_dir.format(root)
            images = [f[:-4] for f in os.listdir(images_dir)
                      if os.path.isfile(os.path.join(images_dir, f)) and f[-3:] == 'jpg']
            ids += [(root, line.strip()) for line in images]
        return ids

    def _load_label(self, idx, height=None, width=None):
        """Parse csv file and return labels."""
        img_id = self._items[idx]
        anno_path = self._anno_path.format(*img_id)
        height = float(height) if height else None
        width = float(width) if width else None
        if idx not in self._im_shapes:
            # store the shapes for later usage
            self._im_shapes[idx] = (width, height)
        with open(anno_path, 'r') as f:
            annotations = f.read().splitlines()
        label = []
        for ann_line in annotations:
            ann_info = ann_line.split(',')
            xmin, ymin = float(ann_info[0]), float(ann_info[1])
            w, h = float(ann_info[2]), float(ann_info[3])
            xmax, ymax = xmin + w, ymin + h
            cls_id = int(ann_info[5]) - 1
            try:
                if not height and not width:
                    self._validate_label(xmin, ymin, xmax, ymax, width, height)
            except AssertionError as e:
                raise RuntimeError("Invalid label at {}, {}".format(anno_path, e))
            label.append([xmin, ymin, xmax, ymax, cls_id])
        return np.array(label)

    def _validate_label(self, xmin, ymin, xmax, ymax, width, height):
        """Validate labels."""
        assert 0 <= xmin < width, "xmin must in [0, {}), given {}".format(width, xmin)
        assert 0 <= ymin < height, "ymin must in [0, {}), given {}".format(height, ymin)
        assert xmin < xmax <= width, "xmax must in (xmin, {}], given {}".format(width, xmax)
        assert ymin < ymax <= height, "ymax must in (ymin, {}], given {}".format(height, ymax)

    def _preload_labels(self):
        """Preload all labels into memory."""
        logging.debug("Preloading %s labels into memory...", str(self))
        return [self._load_label(idx) for idx in range(len(self))]



"""Mask RCNN Estimator."""
# pylint: disable=consider-using-enumerate,abstract-method,arguments-differ
import os
import math
import time
import logging
from multiprocessing import Process

import numpy as np
import mxnet as mx
from mxnet import gluon
from mxnet.contrib import amp

from .... import data as gdata
from .... import utils as gutils
from ....data import COCODetection, VOCDetection
from ....data.transforms import presets
from ....data.transforms.presets.rcnn import MaskRCNNDefaultTrainTransform, MaskRCNNDefaultValTransform
from ....model_zoo import get_model
from ....model_zoo.rcnn.mask_rcnn.data_parallel import ForwardBackwardTask
from ....nn.bbox import BBoxClipToImage
from ....utils.parallel import Parallel
from ....utils.metrics.rcnn import RPNAccMetric, RPNL1LossMetric, RCNNAccMetric, RCNNL1LossMetric, \
    MaskAccMetric, MaskFGAccMetric
from ..base_estimator import BaseEstimator, set_default
from .utils import _get_dataset, _get_dataloader, _save_params, _split_and_load, _get_lr_at_iter

try:
    import horovod.mxnet as hvd
except ImportError:
    hvd = None

try:
    from mpi4py import MPI
except ImportError:
    logging.info('mpi4py is not installed. Use "pip install --no-cache mpi4py" to install')
    MPI = None

from .default import MaskRCNNCfg

__all__ = ['MaskRCNNEstimator']


@set_default(MaskRCNNCfg())
class MaskRCNNEstimator(BaseEstimator):
    """Estimator implementation for Mask-RCNN.

    Parameters
    ----------
    config : dict
        Config in nested dict.
    logger : logging.Logger, default is None
        Optional logger for this estimator, can be `None` when default setting is used.
    reporter : callable, default is None
        If set, use reporter callback to report the metrics of the current estimator.

    Attributes
    ----------
    _logger : logging.Logger
        The customized/default logger for this estimator.
    _logdir : str
        The temporary dir for logs.
    _cfg : autocfg.dataclass
        The configurations.
    """
    def __init__(self, config, logger=None, reporter=None):
        super(MaskRCNNEstimator, self).__init__(config, logger, reporter)

        # fix seed for mxnet, numpy and python builtin random generator.
        gutils.random.seed(self._cfg.train.seed)

        if self._cfg.mask_rcnn.amp:
            amp.init()

        # training contexts
        if self._cfg.horovod:
            self.ctx = [mx.gpu(hvd.local_rank())]
        else:
            ctx = [mx.gpu(int(i)) for i in self._cfg.gpus]
            self.ctx = ctx if ctx else [mx.cpu()]

        # network
        kwargs = {}
        module_list = []
        if self._cfg.mask_rcnn.use_fpn:
            module_list.append('fpn')
        if self._cfg.mask_rcnn.norm_layer is not None:
            module_list.append(self._cfg.mask_rcnn.norm_layer)
            if self._cfg.mask_rcnn.norm_layer == 'bn':
                kwargs['num_devices'] = len(self.ctx)
        self.num_gpus = hvd.size() if self._cfg.horovod else len(self.ctx)
        net_name = '_'.join(('mask_rcnn', *module_list, self._cfg.mask_rcnn.backbone, self._cfg.dataset))
        if self._cfg.mask_rcnn.custom_model:
            self._cfg.mask_rcnn.use_fpn = True
            net_name = '_'.join(('mask_rcnn_fpn', self._cfg.mask_rcnn.backbone, self._cfg.dataset))
            if self._cfg.mask_rcnn.norm_layer == 'bn':
                norm_layer = gluon.contrib.nn.SyncBatchNorm
                norm_kwargs = {'num_devices': len(self.ctx)}
                # sym_norm_layer = mx.sym.contrib.SyncBatchNorm
                sym_norm_kwargs = {'ndev': len(self.ctx)}
            elif self._cfg.mask_rcnn.norm_layer == 'gn':
                norm_layer = gluon.nn.GroupNorm
                norm_kwargs = {'groups': 8}
                # sym_norm_layer = mx.sym.GroupNorm
                sym_norm_kwargs = {'groups': 8}
            else:
                norm_layer = gluon.nn.BatchNorm
                norm_kwargs = None
                # sym_norm_layer = None
                sym_norm_kwargs = None
            if self._cfg.dataset == 'coco':
                classes = COCODetection.CLASSES
            else:
                # default to VOC
                classes = VOCDetection.CLASSES
            self.net = get_model('custom_mask_rcnn_fpn', classes=classes, transfer=None,
                                 dataset=self._cfg.dataset, pretrained_base=self._cfg.train.pretrained_base,
                                 base_network_name=self._cfg.mask_rcnn.backbone, norm_layer=norm_layer,
                                 norm_kwargs=norm_kwargs, sym_norm_kwargs=sym_norm_kwargs,
                                 num_fpn_filters=self._cfg.mask_rcnn.num_fpn_filters,
                                 num_box_head_conv=self._cfg.mask_rcnn.num_box_head_conv,
                                 num_box_head_conv_filters=self._cfg.mask_rcnn.num_box_head_conv_filters,
                                 num_box_head_dense_filters=self._cfg.mask_rcnn.num_box_head_dense_filters,
                                 short=self._cfg.mask_rcnn.image_short, max_size=self._cfg.mask_rcnn.image_max_size,
                                 min_stage=2, max_stage=6, nms_thresh=self._cfg.mask_rcnn.nms_thresh,
                                 nms_topk=self._cfg.mask_rcnn.nms_topk, post_nms=self._cfg.mask_rcnn.post_nms,
                                 roi_mode=self._cfg.mask_rcnn.roi_mode, roi_size=self._cfg.mask_rcnn.roi_size,
                                 strides=self._cfg.mask_rcnn.strides, clip=self._cfg.mask_rcnn.clip,
                                 rpn_channel=self._cfg.mask_rcnn.rpn_channel,
                                 base_size=self._cfg.mask_rcnn.anchor_base_size,
                                 scales=self._cfg.mask_rcnn.anchor_scales,
                                 ratios=self._cfg.mask_rcnn.anchor_aspect_ratio,
                                 alloc_size=self._cfg.mask_rcnn.anchor_alloc_size,
                                 rpn_nms_thresh=self._cfg.mask_rcnn.rpn_nms_thresh,
                                 rpn_train_pre_nms=self._cfg.train.rpn_train_pre_nms,
                                 rpn_train_post_nms=self._cfg.train.rpn_train_post_nms,
                                 rpn_test_pre_nms=self._cfg.valid.rpn_test_pre_nms,
                                 rpn_test_post_nms=self._cfg.valid.rpn_test_post_nms,
                                 rpn_min_size=self._cfg.train.rpn_min_size,
                                 per_device_batch_size=self._cfg.train.batch_size // self.num_gpus,
                                 num_sample=self._cfg.train.rcnn_num_samples,
                                 pos_iou_thresh=self._cfg.train.rcnn_pos_iou_thresh,
                                 pos_ratio=self._cfg.train.rcnn_pos_ratio,
                                 max_num_gt=self._cfg.mask_rcnn.max_num_gt,
                                 target_roi_scale=self._cfg.mask_rcnn.target_roi_scale,
                                 num_fcn_convs=self._cfg.mask_rcnn.num_mask_head_convs)
        else:
            self.net = get_model(net_name, pretrained_base=True,
                                 per_device_batch_size=self._cfg.train.batch_size // self.num_gpus, **kwargs)
        self._cfg.save_prefix += net_name
        if self._cfg.resume.strip():
            self.net.load_parameters(self._cfg.resume.strip())
        else:
            for param in self.net.collect_params().values():
                if param._data is not None:
                    continue
                param.initialize()
        self.net.collect_params().reset_ctx(self.ctx)

        if self._cfg.mask_rcnn.amp:
            # Cast both weights and gradients to 'float16'
            self.net.cast('float16')
            # This layers doesn't support type 'float16'
            self.net.collect_params('.*batchnorm.*').setattr('dtype', 'float32')
            self.net.collect_params('.*normalizedperclassboxcenterencoder.*').setattr('dtype', 'float32')

        # set up logger
        logging.basicConfig()
        self._logger = logging.getLogger()
        self._logger.setLevel(logging.INFO)
        log_file_path = self._cfg.save_prefix + '_train.log'
        log_dir = os.path.dirname(log_file_path)
        if log_dir and not os.path.exists(log_dir):
            os.makedirs(log_dir)
        fh = logging.FileHandler(log_file_path)
        self._logger.addHandler(fh)
        if MPI is None and self._cfg.horovod:
            self._logger.warning('mpi4py is not installed, validation result may be incorrect.')
        self._logger.info(self._cfg)

        self.rpn_cls_loss = mx.gluon.loss.SigmoidBinaryCrossEntropyLoss(from_sigmoid=False)
        self.rpn_box_loss = mx.gluon.loss.HuberLoss(rho=self._cfg.train.rpn_smoothl1_rho)  # == smoothl1
        self.rcnn_cls_loss = mx.gluon.loss.SoftmaxCrossEntropyLoss()
        self.rcnn_box_loss = mx.gluon.loss.HuberLoss(rho=self._cfg.train.rcnn_smoothl1_rho)  # == smoothl1
        self.rcnn_mask_loss = mx.gluon.loss.SigmoidBinaryCrossEntropyLoss(from_sigmoid=False)
        self.metrics = [mx.metric.Loss('RPN_Conf'),
                        mx.metric.Loss('RPN_SmoothL1'),
                        mx.metric.Loss('RCNN_CrossEntropy'),
                        mx.metric.Loss('RCNN_SmoothL1'),
                        mx.metric.Loss('RCNN_Mask')]

        self.rpn_acc_metric = RPNAccMetric()
        self.rpn_bbox_metric = RPNL1LossMetric()
        self.rcnn_acc_metric = RCNNAccMetric()
        self.rcnn_bbox_metric = RCNNL1LossMetric()
        self.rcnn_mask_metric = MaskAccMetric()
        self.rcnn_fgmask_metric = MaskFGAccMetric()
        self.metrics2 = [self.rpn_acc_metric, self.rpn_bbox_metric,
                         self.rcnn_acc_metric, self.rcnn_bbox_metric,
                         self.rcnn_mask_metric, self.rcnn_fgmask_metric]

        self.async_eval_processes = []
        self.best_map = [0]
        self.epoch = 0

        # training data
        self.train_dataset, self.val_dataset, self.eval_metric = _get_dataset(self._cfg.dataset, self._cfg)
        self.batch_size = self._cfg.train.batch_size // self.num_gpus \
            if self._cfg.horovod else self._cfg.train.batch_size
        self._train_data, self._val_data = _get_dataloader(
            self.net, self.train_dataset, self.val_dataset, MaskRCNNDefaultTrainTransform, MaskRCNNDefaultValTransform,
            self.batch_size, len(self.ctx), self._cfg)

    def _validate(self, val_data, async_eval_processes, ctx, eval_metric, logger, epoch, best_map):
        """Test on validation dataset."""
        clipper = BBoxClipToImage()
        eval_metric.reset()
        if not self._cfg.disable_hybridization:
            self.net.hybridize(static_alloc=self._cfg.mask_rcnn.static_alloc)
        tic = time.time()
        for _, batch in enumerate(val_data):
            batch = _split_and_load(batch, ctx_list=ctx)
            det_bboxes = []
            det_ids = []
            det_scores = []
            det_masks = []
            det_infos = []
            for x, im_info in zip(*batch):
                # get prediction results
                ids, scores, bboxes, masks = self.net(x)
                det_bboxes.append(clipper(bboxes, x))
                det_ids.append(ids)
                det_scores.append(scores)
                det_masks.append(masks)
                det_infos.append(im_info)
            # update metric
            for det_bbox, det_id, det_score, det_mask, det_info in zip(det_bboxes, det_ids, det_scores,
                                                                       det_masks, det_infos):
                for i in range(det_info.shape[0]):
                    # numpy everything
                    det_bbox = det_bbox[i].asnumpy()
                    det_id = det_id[i].asnumpy()
                    det_score = det_score[i].asnumpy()
                    det_mask = det_mask[i].asnumpy()
                    det_info = det_info[i].asnumpy()
                    # filter by conf threshold
                    im_height, im_width, im_scale = det_info
                    valid = np.where(((det_id >= 0) & (det_score >= 0.001)))[0]
                    det_id = det_id[valid]
                    det_score = det_score[valid]
                    det_bbox = det_bbox[valid] / im_scale
                    det_mask = det_mask[valid]
                    # fill full mask
                    im_height, im_width = int(round(im_height / im_scale)), int(
                        round(im_width / im_scale))
                    full_masks = gdata.transforms.mask.fill(det_mask, det_bbox, (im_width, im_height))
                    eval_metric.update(det_bbox, det_id, det_score, full_masks)
        if self._cfg.horovod and MPI is not None:
            comm = MPI.COMM_WORLD
            res = comm.gather(eval_metric.get_result_buffer(), root=0)
            if hvd.rank() == 0:
                logger.info('[Epoch {}] Validation Inference cost: {:.3f}'
                            .format(epoch, (time.time() - tic)))
                rank0_res = eval_metric.get_result_buffer()
                if len(rank0_res) == 2:
                    res = res[1:]
                    rank0_res[0].extend([item for res_tuple in res for item in res_tuple[0]])
                    rank0_res[1].extend([item for res_tuple in res for item in res_tuple[1]])
                else:
                    rank0_res.extend([item for r in res for item in r])

        def coco_eval_save_task(eval_metric, logger):
            map_name, mean_ap = eval_metric.get()
            if map_name and mean_ap is not None:
                val_msg = '\n'.join(['{}={}'.format(k, v) for k, v in zip(map_name, mean_ap)])
                logger.info('[Epoch {}] Validation: \n{}'.format(epoch, val_msg))
                current_map = float(mean_ap[-1])
                _save_params(self.net, logger, best_map, current_map, epoch, self._cfg.save_interval,
                             os.path.join(self._logdir, self._cfg.save_prefix))

        if not self._cfg.horovod or hvd.rank() == 0:
            p = Process(target=coco_eval_save_task, args=(eval_metric, self._logger))
            async_eval_processes.append(p)
            p.start()

    def _fit(self, train_data, val_data, time_limit=math.inf):
        """
        Fit Mask R-CNN models.
        """
        # TODO(zhreshold): remove 'dataset' in config, use train_data/val_data instead
        self._cfg.kv_store = 'device' \
            if (self._cfg.mask_rcnn.amp and 'nccl' in self._cfg.kv_store) else self._cfg.kv_store
        kv = mx.kvstore.create(self._cfg.kv_store)
        self.net.collect_params().setattr('grad_req', 'null')
        self.net.collect_train_params().setattr('grad_req', 'write')
        for k, v in self.net.collect_params('.*bias').items():
            v.wd_mult = 0.0
        optimizer_params = {'learning_rate': self._cfg.train.lr, 'wd': self._cfg.train.wd,
                            'momentum': self._cfg.train.momentum, }
        if self._cfg.train.clip_gradient > 0.0:
            optimizer_params['clip_gradient'] = self._cfg.train.clip_gradient
        if self._cfg.mask_rcnn.amp:
            optimizer_params['multi_precision'] = True
        if self._cfg.horovod:
            hvd.broadcast_parameters(self.net.collect_params(), root_rank=0)
            trainer = hvd.DistributedTrainer(
                self.net.collect_train_params(),  # fix batchnorm, fix first stage, etc...
                'sgd',
                optimizer_params
            )
        else:
            trainer = gluon.Trainer(
                self.net.collect_train_params(),  # fix batchnorm, fix first stage, etc...
                'sgd',
                optimizer_params,
                update_on_kvstore=(False if self._cfg.mask_rcnn.amp else None),
                kvstore=kv)

        if self._cfg.mask_rcnn.amp:
            amp.init_trainer(trainer)

        # lr decay policy
        lr_decay = float(self._cfg.train.lr_decay)
        lr_steps = sorted([float(ls) for ls in self._cfg.train.lr_decay_epoch])
        lr_warmup = float(self._cfg.train.lr_warmup)  # avoid int division

        if self._cfg.train.verbose:
            self._logger.info('Trainable parameters:')
            self._logger.info(self.net.collect_train_params().keys())
        self._logger.info('Start training from [Epoch %d]', self._cfg.train.start_epoch)

        base_lr = trainer.learning_rate
        for epoch in range(self._cfg.train.start_epoch, self._cfg.train.epochs):
            self.epoch = epoch
            rcnn_task = ForwardBackwardTask(self.net, trainer, self.rpn_cls_loss, self.rpn_box_loss, self.rcnn_cls_loss,
                                            self.rcnn_box_loss, self.rcnn_mask_loss,
                                            amp_enabled=self._cfg.mask_rcnn.amp)
            executor = Parallel(self._cfg.train.executor_threads, rcnn_task) if not self._cfg.horovod else None
            if not self._cfg.disable_hybridization:
                self.net.hybridize(static_alloc=self._cfg.mask_rcnn.static_alloc)
            while lr_steps and epoch >= lr_steps[0]:
                new_lr = trainer.learning_rate * lr_decay
                lr_steps.pop(0)
                trainer.set_learning_rate(new_lr)
                self._logger.info("[Epoch %d] Set learning rate to %f", epoch, new_lr)
            for metric in self.metrics:
                metric.reset()
            tic = time.time()
            btic = time.time()
            train_data_iter = iter(self._train_data)
            next_data_batch = next(train_data_iter)
            next_data_batch = _split_and_load(next_data_batch, ctx_list=self.ctx)
            for i in range(len(self._train_data)):
                batch = next_data_batch
                if i + epoch * len(self._train_data) <= lr_warmup:
                    # adjust based on real percentage
                    new_lr = base_lr * _get_lr_at_iter((i + epoch * len(self._train_data)) / lr_warmup,
                                                       self._cfg.train.lr_warmup_factor)
                    if new_lr != trainer.learning_rate:
                        if i % self._cfg.train.log_interval == 0:
                            self._logger.info('[Epoch %d Iteration %d] Set learning rate to %f', epoch, i, new_lr)
                        trainer.set_learning_rate(new_lr)
                metric_losses = [[] for _ in self.metrics]
                add_losses = [[] for _ in self.metrics2]
                if executor is not None:
                    for data in zip(*batch):
                        executor.put(data)
                for _ in range(len(self.ctx)):
                    if executor is not None:
                        result = executor.get()
                    else:
                        result = rcnn_task.forward_backward(list(zip(*batch))[0])
                    if (not self._cfg.horovod) or hvd.rank() == 0:
                        for k in range(len(metric_losses)):
                            metric_losses[k].append(result[k])
                        for k in range(len(add_losses)):
                            add_losses[k].append(result[len(metric_losses) + k])
                try:
                    # prefetch next batch
                    next_data_batch = next(train_data_iter)
                    next_data_batch = _split_and_load(next_data_batch, ctx_list=self.ctx)
                except StopIteration:
                    pass

                for metric, record in zip(self.metrics, metric_losses):
                    metric.update(0, record)
                for metric, records in zip(self.metrics2, add_losses):
                    for pred in records:
                        metric.update(pred[0], pred[1])
                trainer.step(self.batch_size)
                if (not self._cfg.horovod or hvd.rank() == 0) and self._cfg.train.log_interval \
                        and not (i + 1) % self._cfg.train.log_interval:
                    msg = ','.join(['{}={:.3f}'.format(*metric.get()) for metric in self.metrics + self.metrics2])
                    self._logger.info('[Epoch {}][Batch {}], Speed: {:.3f} samples/sec, {}'.format(
                        epoch, i,
                        self._cfg.train.log_interval * self._cfg.train.batch_size / (time.time() - btic), msg))
                    btic = time.time()
            # validate and save params
            if (not self._cfg.horovod) or hvd.rank() == 0:
                msg = ','.join(['{}={:.3f}'.format(*metric.get()) for metric in self.metrics])
                self._logger.info('[Epoch {}] Training cost: {:.3f}, {}'.format(
                    epoch, (time.time() - tic), msg))
            if not (epoch + 1) % self._cfg.valid.val_interval:
                # consider reduce the frequency of validation to save time
                self._validate(self._val_data, self.async_eval_processes, self.ctx, self.eval_metric,
                               self._logger, epoch, self.best_map)
            elif (not self._cfg.horovod) or hvd.rank() == 0:
                current_map = 0.
                _save_params(self.net, self._logger, self.best_map, current_map, epoch, self._cfg.save_interval,
                             os.path.join(self._logdir, self._cfg.save_prefix))
            if self._reporter:
                self._reporter(epoch=epoch, map_reward=current_map)
        for thread in self.async_eval_processes:
            thread.join()

    def _evaluate(self, val_data):
        """Evaluate the current model on dataset.
        """
        # TODO(zhreshold): remove self._val_data, use passed in val_data at runtime
        return self._validate(self._val_data, self.async_eval_processes, self.ctx, self.eval_metric,
                              self._logger, self.epoch, self.best_map)

    def predict(self, x):
        """Predict an individual example.

        Parameters
        ----------
        x : file
            An image.
        """
        x, _ = presets.rcnn.transform_test(x, short=self.net.short, max_size=self.net.max_size)
        x = x.as_in_context(self.ctx[0])
        ids, scores, bboxes, masks = [xx[0].asnumpy() for xx in self.net(x)]
        return ids, scores, bboxes, masks



"""Pose utils"""
import time
import logging
import datetime

import numpy as np
import torch

from .. import comm
from ..utils import AverageMeter
from ..model_utils import save_model
from ..optimizer import maybe_add_gradient_clipping
from ..eval_utils.coco_eval import COCOEvaluator

logger = logging.getLogger(__name__)

def _detect_anomaly(losses, loss_dict, iteration):
    if not torch.isfinite(losses).all():
        raise FloatingPointError(
            "Loss became infinite or NaN at iteration={}!\nloss_dict = {}".format(
                iteration, loss_dict
            )
        )


class DirectposePipeline:
    """Train/eval directpose"""
    def __init__(self, base_iter, max_iter, model, dataloader, optimizer, scheduler,
                 cfg, writer=None):
        logger.info('Configuration:\n' + str(cfg))
        self.base_iter = base_iter
        self.max_iter = max_iter
        self.model = model
        self.dataloader = iter(dataloader)
        self.optimizer = optimizer
        self.scheduler = scheduler
        self.cfg = cfg
        self.writer = writer
        self.iter_timer = AverageMeter()
        val_datasets = cfg.CONFIG.DATA.DATASET.VAL
        self.evaluators = [COCOEvaluator(
            val_dataset, cfg, distributed=cfg.DDP_CONFIG.DISTRIBUTED,
            output_dir=cfg.CONFIG.LOG.EVAL_DIR) for val_dataset in val_datasets]

    def train_step(self):
        cfg = self.cfg
        writer = self.writer
        self.model.train()
        end = time.perf_counter()
        data = next(self.dataloader)
        self.base_iter += 1
        if self.base_iter >= self.max_iter:
            return

        data_time = time.perf_counter() - end

        loss_dict = self.model(data)
        losses = sum(loss_dict.values())
        _detect_anomaly(losses, loss_dict, self.base_iter)

        metrics_dict = loss_dict
        metrics_dict["data_time"] = data_time

        self.optimizer.zero_grad()
        losses.backward()
        self.optimizer.step()

        batch_time = time.perf_counter() - end
        end = time.perf_counter()
        metrics_dict["batch_time"] = batch_time

        # gather all metrics
        metrics_dict = {
            k: v.detach().cpu().item() if isinstance(v, torch.Tensor) else float(v)
            for k, v in metrics_dict.items()
        }
        all_metrics_dict = comm.gather(metrics_dict)
        if self.base_iter % cfg.CONFIG.LOG.DISPLAY_FREQ == 0 and cfg.DDP_CONFIG.GPU_WORLD_RANK == 0:
            eta_str = None
            if "data_time" in all_metrics_dict[0]:
                # data_time among workers can have high variance. The actual latency
                # caused by data_time is the maximum among workers.
                data_time = np.max([x.pop("data_time") for x in all_metrics_dict])
            if "batch_time" in all_metrics_dict[0]:
                # batch_time among workers can have high variance. The actual latency
                # caused by batch_time is the maximum among workers.
                batch_time = np.max([x.pop("batch_time") for x in all_metrics_dict])
                self.iter_timer.update(batch_time)
                eta = (self.max_iter - self.base_iter) * self.iter_timer.avg
                eta_str = str(datetime.timedelta(seconds=int(eta)))
            # average the rest metrics
            metrics_dict = {
                k: np.mean([x[k] for x in all_metrics_dict]) for k in all_metrics_dict[0].keys()
            }
            total_losses_reduced = sum(loss for loss in metrics_dict.values())
            for param in self.optimizer.param_groups:
                lr = param['lr']
            print_string = 'Iter: [{0}/{1}]'.format(
                self.base_iter, self.max_iter)
            if eta_str is not None:
                print_string += f' ETA: {eta_str} '
            print_string += ' data_time: {data_time:.3f}, batch time: {batch_time:.3f}'.format(
                data_time=data_time, batch_time=batch_time)
            print_string += ' loss: {loss:.5f}'.format(loss=total_losses_reduced)
            print_string += ' lr: {lr:.6f}'.format(lr=lr)
            iteration = self.base_iter
            writer.add_scalar('total_loss', total_losses_reduced, iteration)
            writer.add_scalar('learning_rate', lr, iteration)
            if len(metrics_dict) > 1:
                for km, kv in metrics_dict.items():
                    writer.add_scalar(km, kv, iteration)
                    print_string += f' {km}: {kv:.2f}'
            logger.info(print_string)
        self.scheduler.step()

    def validate(self, val_loader):
        if not self.evaluators:
            self.info('No evaluation data specified, skip...')
            return
        # save model before evaluation so we may recover from bugs
        self.save_model()
        is_training = self.model.training
        self.model.eval()
        for evaluator in self.evaluators:
            evaluator.reset()

        # inference
        num_devices = comm.get_world_size()
        total = len(val_loader)
        logger.info(f"Start evaluation on {total} images.")
        warmup_meter = min(5, total - 1)
        start_time = time.perf_counter()
        total_compute_time = 0
        with torch.no_grad():
            for idx, inputs in enumerate(val_loader):
                if idx <= warmup_meter:
                    # no timing during gpu warm up
                    start_time = time.perf_counter()
                    total_compute_time = 0

                start_compute_time = time.perf_counter()
                outputs = self.model(inputs)
                if torch.cuda.is_available():
                    torch.cuda.synchronize()
                total_compute_time += time.perf_counter() - start_compute_time
                for evaluator in self.evaluators:
                    evaluator.process(inputs, outputs)
                iters_after_start = idx + 1 - warmup_meter * int(idx >= warmup_meter)
                seconds_per_img = total_compute_time / iters_after_start
                if idx >= warmup_meter * 2 or seconds_per_img > 5:
                    total_seconds_per_img = (time.perf_counter() - start_time) / iters_after_start
                    eta = datetime.timedelta(seconds=int(total_seconds_per_img * (total - idx - 1)))
                    logger.info(f"Inference done {idx + 1}/{total}. {seconds_per_img:.4f} sec/img. ETA={str(eta)}")

        # Measure the time only for this worker (before the synchronization barrier)
        total_time = time.perf_counter() - start_time
        total_time_str = str(datetime.timedelta(seconds=total_time))
        # NOTE this format is parsed by grep
        logger.info(
            "Total inference time: {} ({:.6f} s / img per device, on {} devices)".format(
                total_time_str, total_time / (total - warmup_meter), num_devices
            )
        )
        total_compute_time_str = str(datetime.timedelta(seconds=int(total_compute_time)))
        logger.info(
            "Total inference pure compute time: {} ({:.6f} s / img per device, on {} devices)".format(
                total_compute_time_str, total_compute_time / (total - warmup_meter), num_devices
            )
        )
        for evaluator in self.evaluators:
            results = evaluator.evaluate()
            if self.cfg.DDP_CONFIG.GPU_WORLD_RANK == 0:
                assert isinstance(results, dict)
                for task, res in results.items():
                    # Don't print "AP-category" metrics since they are usually not tracked.
                    important_res = [(k, v) for k, v in res.items() if "-" not in k]
                    logger.info("copypaste: Task: {}".format(task))
                    logger.info("copypaste: " + ",".join([k[0] for k in important_res]))
                    logger.info("copypaste: " + ",".join(["{0:.4f}".format(k[1]) for k in important_res]))
                    for k in important_res:
                        self.writer.add_scalar(':'.join((task, k[0])), k[1], self.base_iter)

        # resume training state if applicable
        self.model.train(is_training)

    def save_model(self):
        """Save checkpoint"""
        if self.cfg.DDP_CONFIG.GPU_WORLD_RANK == 0:
            save_model(self.model, self.optimizer, self.base_iter, self.cfg)


def build_pose_optimizer(cfg, model) -> torch.optim.Optimizer:
    """
    Build an optimizer from config.
    """
    norm_module_types = (
        torch.nn.BatchNorm1d,
        torch.nn.BatchNorm2d,
        torch.nn.BatchNorm3d,
        torch.nn.SyncBatchNorm,
        # NaiveSyncBatchNorm inherits from BatchNorm2d
        torch.nn.GroupNorm,
        torch.nn.InstanceNorm1d,
        torch.nn.InstanceNorm2d,
        torch.nn.InstanceNorm3d,
        torch.nn.LayerNorm,
        torch.nn.LocalResponseNorm,
    )
    params: List[Dict[str, Any]] = []
    memo: Set[torch.nn.parameter.Parameter] = set()
    override: Set[torch.nn.parameter.Parameter] = set()

    for module in model.modules():
        for key, value in module.named_parameters(recurse=False):
            if not value.requires_grad:
                continue
            # Avoid duplicating parameters
            if value in memo:
                continue
            memo.add(value)
            lr = cfg.CONFIG.TRAIN.LR
            weight_decay = cfg.CONFIG.TRAIN.W_DECAY
            if isinstance(module, norm_module_types):
                weight_decay = cfg.CONFIG.TRAIN.W_DECAY_NORM
            elif key == "bias":
                lr = cfg.CONFIG.TRAIN.LR
                weight_decay = cfg.CONFIG.TRAIN.W_DECAY
            if value in override:
                raise NotImplementedError('KPS_GRAD_MULT not found')

            params += [{"params": [value], "lr": lr, "weight_decay": weight_decay}]

    optimizer = torch.optim.SGD(
        params, cfg.CONFIG.TRAIN.LR, momentum=cfg.CONFIG.TRAIN.MOMENTUM)
    optimizer = maybe_add_gradient_clipping(cfg, optimizer)
    return optimizer



# pylint: disable=missing-function-docstring, line-too-long
"""
SlowFast Networks for Video Recognition
ICCV 2019, https://arxiv.org/abs/1812.03982
Code adapted from https://github.com/open-mmlab/mmaction and
https://github.com/decisionforce/TPN
"""
import torch
import torch.nn as nn
import torch.utils.checkpoint as cp

from .non_local import build_nonlocal_block


__all__ = ['ResNet_SlowFast', 'i3d_slow_resnet50_f32s2_kinetics400',
           'i3d_slow_resnet50_f16s4_kinetics400', 'i3d_slow_resnet50_f8s8_kinetics400',
           'i3d_slow_resnet101_f32s2_kinetics400', 'i3d_slow_resnet101_f16s4_kinetics400',
           'i3d_slow_resnet101_f8s8_kinetics400', 'i3d_slow_resnet50_f32s2_custom', 'i3d_slow_resnet101_f16s4_kinetics700']


def conv3x3x3(in_planes, out_planes, spatial_stride=1, temporal_stride=1, dilation=1):
    "3x3x3 convolution with padding"
    return nn.Conv3d(in_planes,
                     out_planes,
                     kernel_size=3,
                     stride=(temporal_stride, spatial_stride, spatial_stride),
                     padding=dilation,
                     dilation=dilation,
                     bias=False)


def conv1x3x3(in_planes, out_planes, spatial_stride=1, temporal_stride=1, dilation=1):
    "1x3x3 convolution with padding"
    return nn.Conv3d(in_planes,
                     out_planes,
                     kernel_size=(1, 3, 3),
                     stride=(temporal_stride, spatial_stride, spatial_stride),
                     padding=(0, dilation, dilation),
                     dilation=dilation,
                     bias=False)


class Bottleneck(nn.Module):
    """Bottleneck block for ResNet.
    If style is "pytorch", the stride-two layer is the 3x3 conv layer,
    if it is "caffe", the stride-two layer is the first 1x1 conv layer.
    """
    expansion = 4

    def __init__(self,
                 inplanes,
                 planes,
                 spatial_stride=1,
                 temporal_stride=1,
                 dilation=1,
                 downsample=None,
                 style='pytorch',
                 if_inflate=True,
                 inflate_style='3x1x1',
                 if_nonlocal=True,
                 nonlocal_cfg=None,
                 with_cp=False):

        super(Bottleneck, self).__init__()
        assert style in ['pytorch', 'caffe']
        assert inflate_style in ['3x1x1', '3x3x3']
        self.inplanes = inplanes
        self.planes = planes

        if style == 'pytorch':
            self.conv1_stride = 1
            self.conv2_stride = spatial_stride
            self.conv1_stride_t = 1
            self.conv2_stride_t = temporal_stride
        else:
            self.conv1_stride = spatial_stride
            self.conv2_stride = 1
            self.conv1_stride_t = temporal_stride
            self.conv2_stride_t = 1
        if if_inflate:
            if inflate_style == '3x1x1':
                self.conv1 = nn.Conv3d(
                    inplanes,
                    planes,
                    kernel_size=(3, 1, 1),
                    stride=(self.conv1_stride_t, self.conv1_stride, self.conv1_stride),
                    padding=(1, 0, 0),
                    bias=False)
                self.conv2 = nn.Conv3d(
                    planes,
                    planes,
                    kernel_size=(1, 3, 3),
                    stride=(self.conv2_stride_t, self.conv2_stride, self.conv2_stride),
                    padding=(0, dilation, dilation),
                    dilation=(1, dilation, dilation),
                    bias=False)
            else:
                self.conv1 = nn.Conv3d(
                    inplanes,
                    planes,
                    kernel_size=1,
                    stride=(self.conv1_stride_t, self.conv1_stride, self.conv1_stride),
                    bias=False)
                self.conv2 = nn.Conv3d(
                    planes,
                    planes,
                    kernel_size=3,
                    stride=(self.conv2_stride_t, self.conv2_stride, self.conv2_stride),
                    padding=(1, dilation, dilation),
                    dilation=(1, dilation, dilation),
                    bias=False)
        else:
            self.conv1 = nn.Conv3d(
                inplanes,
                planes,
                kernel_size=1,
                stride=(1, self.conv1_stride, self.conv1_stride),
                bias=False)
            self.conv2 = nn.Conv3d(
                planes,
                planes,
                kernel_size=(1, 3, 3),
                stride=(1, self.conv2_stride, self.conv2_stride),
                padding=(0, dilation, dilation),
                dilation=(1, dilation, dilation),
                bias=False)

        self.bn1 = nn.BatchNorm3d(planes)
        self.bn2 = nn.BatchNorm3d(planes)
        self.conv3 = nn.Conv3d(
            planes, planes * self.expansion, kernel_size=1, bias=False)
        self.bn3 = nn.BatchNorm3d(planes * self.expansion)
        self.relu = nn.ReLU(inplace=True)
        self.downsample = downsample
        self.spatial_tride = spatial_stride
        self.temporal_tride = temporal_stride
        self.dilation = dilation
        self.with_cp = with_cp

        if if_nonlocal and nonlocal_cfg is not None:
            nonlocal_cfg_ = nonlocal_cfg.copy()
            nonlocal_cfg_['in_channels'] = planes * self.expansion
            self.nonlocal_block = build_nonlocal_block(nonlocal_cfg_)
        else:
            self.nonlocal_block = None

    def forward(self, x):

        def _inner_forward(x):
            identity = x

            out = self.conv1(x)
            out = self.bn1(out)
            out = self.relu(out)

            out = self.conv2(out)
            out = self.bn2(out)
            out = self.relu(out)

            out = self.conv3(out)
            out = self.bn3(out)

            if self.downsample is not None:
                identity = self.downsample(x)

            out += identity

            return out

        if self.with_cp and x.requires_grad:
            out = cp.checkpoint(_inner_forward, x)
        else:
            out = _inner_forward(x)

        out = self.relu(out)

        if self.nonlocal_block is not None:
            out = self.nonlocal_block(out)

        return out


def make_res_layer(block,
                   inplanes,
                   planes,
                   blocks,
                   spatial_stride=1,
                   temporal_stride=1,
                   dilation=1,
                   style='pytorch',
                   inflate_freq=1,
                   inflate_style='3x1x1',
                   nonlocal_freq=1,
                   nonlocal_cfg=None,
                   with_cp=False):
    inflate_freq = inflate_freq if not isinstance(inflate_freq, int) else (inflate_freq,) * blocks
    nonlocal_freq = nonlocal_freq if not isinstance(nonlocal_freq, int) else (nonlocal_freq,) * blocks
    assert len(inflate_freq) == blocks
    assert len(nonlocal_freq) == blocks
    downsample = None
    if spatial_stride != 1 or inplanes != planes * block.expansion:
        downsample = nn.Sequential(
            nn.Conv3d(
                inplanes,
                planes * block.expansion,
                kernel_size=1,
                stride=(temporal_stride, spatial_stride, spatial_stride),
                bias=False),
            nn.BatchNorm3d(planes * block.expansion),
        )

    layers = []
    layers.append(
        block(
            inplanes,
            planes,
            spatial_stride,
            temporal_stride,
            dilation,
            downsample,
            style=style,
            if_inflate=(inflate_freq[0] == 1),
            inflate_style=inflate_style,
            if_nonlocal=(nonlocal_freq[0] == 1),
            nonlocal_cfg=nonlocal_cfg,
            with_cp=with_cp))
    inplanes = planes * block.expansion
    for i in range(1, blocks):
        layers.append(
            block(inplanes,
                  planes,
                  1, 1,
                  dilation,
                  style=style,
                  if_inflate=(inflate_freq[i] == 1),
                  inflate_style=inflate_style,
                  if_nonlocal=(nonlocal_freq[i] == 1),
                  nonlocal_cfg=nonlocal_cfg,
                  with_cp=with_cp))

    return nn.Sequential(*layers)


class ResNet_SlowFast(nn.Module):
    """ResNe(x)t_SlowFast backbone.
    Args:
        depth (int): Depth of resnet, from {50, 101, 152}.
        num_stages (int): Resnet stages, normally 4.
        strides (Sequence[int]): Strides of the first block of each stage.
        dilations (Sequence[int]): Dilation of each stage.
        out_indices (Sequence[int]): Output from which stages.
        style (str): `pytorch` or `caffe`. If set to "pytorch", the stride-two
            layer is the 3x3 conv layer, otherwise the stride-two layer is
            the first 1x1 conv layer.
        frozen_stages (int): Stages to be frozen (all param fixed). -1 means
            not freezing any parameters.
        bn_eval (bool): Whether to set BN layers to eval mode, namely, freeze
            running stats (mean and var).
        bn_frozen (bool): Whether to freeze weight and bias of BN layers.
        with_cp (bool): Use checkpoint or not. Using checkpoint will save some
            memory while slowing down the training speed.
    """

    arch_settings = {
        50: (Bottleneck, (3, 4, 6, 3)),
        101: (Bottleneck, (3, 4, 23, 3)),
        152: (Bottleneck, (3, 8, 36, 3))
    }

    def __init__(self,
                 num_classes,
                 depth,
                 pretrained=None,
                 pretrained_base=True,
                 feat_ext=False,
                 num_stages=4,
                 spatial_strides=(1, 2, 2, 2),
                 temporal_strides=(1, 1, 1, 1),
                 dilations=(1, 1, 1, 1),
                 out_indices=(0, 1, 2, 3),
                 conv1_kernel_t=1,
                 conv1_stride_t=1,
                 pool1_kernel_t=1,
                 pool1_stride_t=1,
                 style='pytorch',
                 frozen_stages=-1,
                 inflate_freq=(0, 0, 1, 1),
                 inflate_stride=(1, 1, 1, 1),
                 inflate_style='3x1x1',
                 nonlocal_stages=(-1,),
                 nonlocal_freq=(0, 0, 0, 0),
                 nonlocal_cfg=None,
                 bn_eval=False,
                 bn_frozen=False,
                 partial_bn=False,
                 with_cp=False,
                 dropout_ratio=0.5,
                 init_std=0.01):
        super(ResNet_SlowFast, self).__init__()

        if depth not in self.arch_settings:
            raise KeyError('invalid depth {} for resnet'.format(depth))

        self.num_classes = num_classes
        self.depth = depth
        self.pretrained = pretrained
        self.pretrained_base = pretrained_base
        self.num_stages = num_stages
        assert 1 <= num_stages <= 4
        self.spatial_strides = spatial_strides
        self.temporal_strides = temporal_strides
        self.dilations = dilations
        assert len(spatial_strides) == len(temporal_strides) == len(dilations) == num_stages
        self.out_indices = out_indices
        assert max(out_indices) < num_stages
        self.style = style
        self.frozen_stages = frozen_stages
        self.inflate_freqs = inflate_freq if not isinstance(inflate_freq, int) else (inflate_freq,) * num_stages
        self.inflate_style = inflate_style
        self.nonlocal_stages = nonlocal_stages
        self.nonlocal_freqs = nonlocal_freq if not isinstance(nonlocal_freq, int) else (nonlocal_freq,) * num_stages
        self.nonlocal_cfg = nonlocal_cfg
        self.bn_eval = bn_eval
        self.bn_frozen = bn_frozen
        self.partial_bn = partial_bn
        self.with_cp = with_cp
        self.feat_ext = feat_ext

        self.dropout_ratio = dropout_ratio
        self.init_std = init_std

        self.block, stage_blocks = self.arch_settings[depth]
        self.stage_blocks = stage_blocks[:num_stages]
        self.inplanes = 64

        self.conv1 = nn.Conv3d(
            3, 64, kernel_size=(conv1_kernel_t, 7, 7), stride=(conv1_stride_t, 2, 2),
            padding=((conv1_kernel_t - 1) // 2, 3, 3), bias=False)
        self.bn1 = nn.BatchNorm3d(64)
        self.relu = nn.ReLU(inplace=True)
        self.maxpool = nn.MaxPool3d(kernel_size=(pool1_kernel_t, 3, 3), stride=(pool1_stride_t, 2, 2),
                                    padding=(pool1_kernel_t // 2, 1, 1))

        self.res_layers = []
        for i, num_blocks in enumerate(self.stage_blocks):
            spatial_stride = spatial_strides[i]
            temporal_stride = temporal_strides[i]
            dilation = dilations[i]
            planes = 64 * 2 ** i
            res_layer = make_res_layer(
                self.block,
                self.inplanes,
                planes,
                num_blocks,
                spatial_stride=spatial_stride,
                temporal_stride=temporal_stride,
                dilation=dilation,
                style=self.style,
                inflate_freq=self.inflate_freqs[i],
                inflate_style=self.inflate_style,
                nonlocal_freq=self.nonlocal_freqs[i],
                nonlocal_cfg=self.nonlocal_cfg if i in self.nonlocal_stages else None,
                with_cp=with_cp)
            self.inplanes = planes * self.block.expansion
            layer_name = 'layer{}'.format(i + 1)
            self.add_module(layer_name, res_layer)
            self.res_layers.append(layer_name)

        self.feat_dim = self.block.expansion * 64 * 2 ** (len(self.stage_blocks) - 1)

        if self.dropout_ratio != 0:
            self.dropout = nn.Dropout(p=self.dropout_ratio)
        else:
            self.dropout = None

        self.avg_pool = nn.AdaptiveAvgPool3d(1)
        self.fc = nn.Linear(in_features=2048, out_features=num_classes)
        if not self.pretrained:
            nn.init.normal_(self.fc.weight, 0, self.init_std)
            nn.init.constant_(self.fc.bias, 0)

    def forward(self, x):
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)
        x = self.maxpool(x)

        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.layer4(x)

        x = self.avg_pool(x)
        if self.dropout is not None:
            x = self.dropout(x)
        x = x.view(x.size(0), -1)

        if self.feat_ext:
            return x

        out = self.fc(x)
        return out


def i3d_slow_resnet50_f32s2_kinetics400(cfg):
    model = ResNet_SlowFast(num_classes=cfg.CONFIG.DATA.NUM_CLASSES,
                            depth=50,
                            pretrained=cfg.CONFIG.MODEL.PRETRAINED,
                            pretrained_base=cfg.CONFIG.MODEL.PRETRAINED_BASE,
                            feat_ext=cfg.CONFIG.INFERENCE.FEAT,
                            bn_eval=cfg.CONFIG.MODEL.BN_EVAL,
                            partial_bn=cfg.CONFIG.MODEL.PARTIAL_BN,
                            bn_frozen=cfg.CONFIG.MODEL.BN_FROZEN)

    if cfg.CONFIG.MODEL.PRETRAINED:
        from ..model_store import get_model_file
        model.load_state_dict(torch.load(get_model_file('i3d_slow_resnet50_f32s2_kinetics400',
                                                        tag=cfg.CONFIG.MODEL.PRETRAINED)))
    return model


def i3d_slow_resnet50_f16s4_kinetics400(cfg):
    model = ResNet_SlowFast(num_classes=cfg.CONFIG.DATA.NUM_CLASSES,
                            depth=50,
                            pretrained=cfg.CONFIG.MODEL.PRETRAINED,
                            pretrained_base=cfg.CONFIG.MODEL.PRETRAINED_BASE,
                            feat_ext=cfg.CONFIG.INFERENCE.FEAT,
                            bn_eval=cfg.CONFIG.MODEL.BN_EVAL,
                            partial_bn=cfg.CONFIG.MODEL.PARTIAL_BN,
                            bn_frozen=cfg.CONFIG.MODEL.BN_FROZEN)

    if cfg.CONFIG.MODEL.PRETRAINED:
        from ..model_store import get_model_file
        model.load_state_dict(torch.load(get_model_file('i3d_slow_resnet50_f16s4_kinetics400',
                                                        tag=cfg.CONFIG.MODEL.PRETRAINED)))
    return model


def i3d_slow_resnet50_f8s8_kinetics400(cfg):
    model = ResNet_SlowFast(num_classes=cfg.CONFIG.DATA.NUM_CLASSES,
                            depth=50,
                            pretrained=cfg.CONFIG.MODEL.PRETRAINED,
                            pretrained_base=cfg.CONFIG.MODEL.PRETRAINED_BASE,
                            feat_ext=cfg.CONFIG.INFERENCE.FEAT,
                            bn_eval=cfg.CONFIG.MODEL.BN_EVAL,
                            partial_bn=cfg.CONFIG.MODEL.PARTIAL_BN,
                            bn_frozen=cfg.CONFIG.MODEL.BN_FROZEN)

    if cfg.CONFIG.MODEL.PRETRAINED:
        from ..model_store import get_model_file
        model.load_state_dict(torch.load(get_model_file('i3d_slow_resnet50_f8s8_kinetics400',
                                                        tag=cfg.CONFIG.MODEL.PRETRAINED)))
    return model


def i3d_slow_resnet101_f32s2_kinetics400(cfg):
    model = ResNet_SlowFast(num_classes=cfg.CONFIG.DATA.NUM_CLASSES,
                            depth=101,
                            pretrained=cfg.CONFIG.MODEL.PRETRAINED,
                            pretrained_base=cfg.CONFIG.MODEL.PRETRAINED_BASE,
                            feat_ext=cfg.CONFIG.INFERENCE.FEAT,
                            bn_eval=cfg.CONFIG.MODEL.BN_EVAL,
                            partial_bn=cfg.CONFIG.MODEL.PARTIAL_BN,
                            bn_frozen=cfg.CONFIG.MODEL.BN_FROZEN)

    if cfg.CONFIG.MODEL.PRETRAINED:
        from ..model_store import get_model_file
        model.load_state_dict(torch.load(get_model_file('i3d_slow_resnet101_f32s2_kinetics400',
                                                        tag=cfg.CONFIG.MODEL.PRETRAINED)))
    return model


def i3d_slow_resnet101_f16s4_kinetics400(cfg):
    model = ResNet_SlowFast(num_classes=cfg.CONFIG.DATA.NUM_CLASSES,
                            depth=101,
                            pretrained=cfg.CONFIG.MODEL.PRETRAINED,
                            pretrained_base=cfg.CONFIG.MODEL.PRETRAINED_BASE,
                            feat_ext=cfg.CONFIG.INFERENCE.FEAT,
                            bn_eval=cfg.CONFIG.MODEL.BN_EVAL,
                            partial_bn=cfg.CONFIG.MODEL.PARTIAL_BN,
                            bn_frozen=cfg.CONFIG.MODEL.BN_FROZEN)

    if cfg.CONFIG.MODEL.PRETRAINED:
        from ..model_store import get_model_file
        model.load_state_dict(torch.load(get_model_file('i3d_slow_resnet101_f16s4_kinetics400',
                                                        tag=cfg.CONFIG.MODEL.PRETRAINED)))
    return model

def i3d_slow_resnet101_f16s4_kinetics700(cfg):
    model = ResNet_SlowFast(num_classes=cfg.CONFIG.DATA.NUM_CLASSES,
                            depth=101,
                            pretrained=cfg.CONFIG.MODEL.PRETRAINED,
                            pretrained_base=cfg.CONFIG.MODEL.PRETRAINED_BASE,
                            feat_ext=cfg.CONFIG.INFERENCE.FEAT,
                            bn_eval=cfg.CONFIG.MODEL.BN_EVAL,
                            partial_bn=cfg.CONFIG.MODEL.PARTIAL_BN,
                            bn_frozen=cfg.CONFIG.MODEL.BN_FROZEN)

    if cfg.CONFIG.MODEL.PRETRAINED:
        from ..model_store import get_model_file
        model.load_state_dict(torch.load(get_model_file('i3d_slow_resnet101_f16s4_kinetics700',
                                                        tag=cfg.CONFIG.MODEL.PRETRAINED)))
    return model

def i3d_slow_resnet101_f8s8_kinetics400(cfg):
    model = ResNet_SlowFast(num_classes=cfg.CONFIG.DATA.NUM_CLASSES,
                            depth=101,
                            pretrained=cfg.CONFIG.MODEL.PRETRAINED,
                            pretrained_base=cfg.CONFIG.MODEL.PRETRAINED_BASE,
                            feat_ext=cfg.CONFIG.INFERENCE.FEAT,
                            bn_eval=cfg.CONFIG.MODEL.BN_EVAL,
                            partial_bn=cfg.CONFIG.MODEL.PARTIAL_BN,
                            bn_frozen=cfg.CONFIG.MODEL.BN_FROZEN)

    if cfg.CONFIG.MODEL.PRETRAINED:
        from ..model_store import get_model_file
        model.load_state_dict(torch.load(get_model_file('i3d_slow_resnet101_f8s8_kinetics400',
                                                        tag=cfg.CONFIG.MODEL.PRETRAINED)))
    return model


def i3d_slow_resnet50_f32s2_custom(cfg):
    model = ResNet_SlowFast(num_classes=cfg.CONFIG.DATA.NUM_CLASSES,
                            depth=50,
                            pretrained=cfg.CONFIG.MODEL.PRETRAINED,
                            pretrained_base=cfg.CONFIG.MODEL.PRETRAINED_BASE,
                            feat_ext=cfg.CONFIG.INFERENCE.FEAT,
                            bn_eval=cfg.CONFIG.MODEL.BN_EVAL,
                            partial_bn=cfg.CONFIG.MODEL.PARTIAL_BN,
                            bn_frozen=cfg.CONFIG.MODEL.BN_FROZEN)

    if cfg.CONFIG.MODEL.PRETRAINED:
        from ..model_store import get_model_file
        state_dict = torch.load(get_model_file('i3d_slow_resnet50_f32s2_kinetics400', tag=cfg.CONFIG.MODEL.PRETRAINED))
        for k in list(state_dict.keys()):
            # retain only backbone up to before the classification layer
            if k.startswith('fc'):
                del state_dict[k]

        msg = model.load_state_dict(state_dict, strict=False)
        assert set(msg.missing_keys) == {'fc.weight', 'fc.bias'}
        print("=> Initialized from a I3D_slow model pretrained on Kinetcis400 dataset")
    return model



# pylint: disable=missing-function-docstring, missing-class-docstring, unused-argument, line-too-long
"""
PyTorch implementation of the I3D model.
Quo Vadis, Action Recognition? A New Model and the Kinetics Dataset
CVPR 2017, https://arxiv.org/abs/1705.07750
"""
import torch
import torch.nn as nn
from torch.nn import BatchNorm3d

from .non_local import build_nonlocal_block


__all__ = ['I3D_ResNetV1', 'i3d_resnet50_v1_kinetics400', 'i3d_resnet101_v1_kinetics400',
           'i3d_nl5_resnet50_v1_kinetics400', 'i3d_nl10_resnet50_v1_kinetics400',
           'i3d_nl5_resnet101_v1_kinetics400', 'i3d_nl10_resnet101_v1_kinetics400',
           'i3d_resnet50_v1_sthsthv2', 'i3d_resnet50_v1_custom']


def conv3x3x3(in_planes, out_planes, spatial_stride=1, temporal_stride=1, dilation=1):
    "3x3x3 convolution with padding"
    return nn.Conv3d(in_channels=in_planes,
                     out_channels=out_planes,
                     kernel_size=3,
                     stride=(temporal_stride, spatial_stride, spatial_stride),
                     dilation=dilation,
                     bias=False)


def conv1x3x3(in_planes, out_planes, spatial_stride=1, temporal_stride=1, dilation=1):
    "1x3x3 convolution with padding"
    return nn.Conv3d(in_channels=in_planes,
                     out_channels=out_planes,
                     kernel_size=(1, 3, 3),
                     stride=(temporal_stride, spatial_stride, spatial_stride),
                     padding=(0, dilation, dilation),
                     dilation=dilation,
                     bias=False)


class BasicBlock(nn.Module):
    expansion = 1

    def __init__(self,
                 inplanes,
                 planes,
                 spatial_stride=1,
                 temporal_stride=1,
                 dilation=1,
                 downsample=None,
                 if_inflate=True,
                 inflate_style=None,
                 norm_layer=BatchNorm3d,
                 norm_kwargs=None,
                 layer_name='',
                 **kwargs):
        super(BasicBlock, self).__init__()

        if if_inflate:
            self.conv1 = conv3x3x3(inplanes, planes, spatial_stride, temporal_stride, dilation)
        else:
            self.conv1 = conv1x3x3(inplanes, planes, spatial_stride, temporal_stride, dilation)
        self.bn1 = norm_layer(**({} if norm_kwargs is None else norm_kwargs))
        self.relu = nn.ReLU(inplace=False)
        if if_inflate:
            self.conv2 = conv3x3x3(planes, planes)
        else:
            self.conv2 = conv1x3x3(planes, planes)
        self.bn2 = norm_layer(**({} if norm_kwargs is None else norm_kwargs))

        self.downsample = downsample
        self.spatial_stride = spatial_stride
        self.temporal_stride = temporal_stride
        self.dilation = dilation

    def forward(self, x):
        identity = x
        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)
        out = self.conv2(out)
        out = self.bn2(out)
        if self.downsample is not None:
            identity = self.downsample(x)
        out = out + identity
        out = self.relu(out)
        return out


class Bottleneck(nn.Module):
    expansion = 4

    def __init__(self,
                 inplanes,
                 planes,
                 spatial_stride=1,
                 temporal_stride=1,
                 dilation=1,
                 downsample=None,
                 if_inflate=True,
                 inflate_style='3x1x1',
                 if_nonlocal=True,
                 nonlocal_cfg=None,
                 norm_layer=BatchNorm3d,
                 norm_kwargs=None,
                 layer_name='',
                 **kwargs):
        super(Bottleneck, self).__init__()
        assert inflate_style in ['3x1x1', '3x3x3']
        self.inplanes = inplanes
        self.planes = planes

        self.conv1_stride = 1
        self.conv2_stride = spatial_stride
        self.conv1_stride_t = 1
        self.conv2_stride_t = temporal_stride

        if if_inflate:
            if inflate_style == '3x1x1':
                self.conv1 = nn.Conv3d(in_channels=inplanes,
                                       out_channels=planes,
                                       kernel_size=(3, 1, 1),
                                       stride=(self.conv1_stride_t, self.conv1_stride, self.conv1_stride),
                                       padding=(1, 0, 0),
                                       bias=False)
                self.conv2 = nn.Conv3d(in_channels=planes,
                                       out_channels=planes,
                                       kernel_size=(1, 3, 3),
                                       stride=(self.conv2_stride_t, self.conv2_stride, self.conv2_stride),
                                       padding=(0, dilation, dilation),
                                       dilation=(1, dilation, dilation),
                                       bias=False)
            else:
                self.conv1 = nn.Conv3d(in_channels=inplanes,
                                       out_channels=planes,
                                       kernel_size=1,
                                       stride=(self.conv1_stride_t, self.conv1_stride, self.conv1_stride),
                                       bias=False)
                self.conv2 = nn.Conv3d(in_channels=planes,
                                       out_channels=planes,
                                       kernel_size=3,
                                       stride=(self.conv2_stride_t, self.conv2_stride, self.conv2_stride),
                                       padding=(1, dilation, dilation),
                                       dilation=(1, dilation, dilation),
                                       bias=False)
        else:
            self.conv1 = nn.Conv3d(in_channels=inplanes,
                                   out_channels=planes,
                                   kernel_size=1,
                                   stride=(1, self.conv1_stride, self.conv1_stride),
                                   bias=False)
            self.conv2 = nn.Conv3d(in_channels=planes,
                                   out_channels=planes,
                                   kernel_size=(1, 3, 3),
                                   stride=(1, self.conv2_stride, self.conv2_stride),
                                   padding=(0, dilation, dilation),
                                   dilation=(1, dilation, dilation),
                                   bias=False)

        self.bn1 = norm_layer(num_features=planes, **({} if norm_kwargs is None else norm_kwargs))
        self.bn2 = norm_layer(num_features=planes, **({} if norm_kwargs is None else norm_kwargs))
        self.conv3 = nn.Conv3d(in_channels=planes,
                               out_channels=planes * self.expansion,
                               kernel_size=1,
                               bias=False)
        self.bn3 = norm_layer(num_features=planes * self.expansion, **({} if norm_kwargs is None else norm_kwargs))
        self.downsample = downsample

        self.spatial_tride = spatial_stride
        self.temporal_tride = temporal_stride
        self.dilation = dilation

        if if_nonlocal and nonlocal_cfg is not None:
            nonlocal_cfg_ = nonlocal_cfg.copy()
            nonlocal_cfg_['in_channels'] = planes * self.expansion
            self.nonlocal_block = build_nonlocal_block(nonlocal_cfg_)
        else:
            self.nonlocal_block = None

        self.relu = nn.ReLU(inplace=False)

    def forward(self, x):
        identity = x
        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)
        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)
        out = self.conv3(out)
        out = self.bn3(out)
        if self.downsample is not None:
            identity = self.downsample(x)
        out = out + identity
        out = self.relu(out)

        if self.nonlocal_block is not None:
            out = self.nonlocal_block(out)
        return out


def make_res_layer(block,
                   inplanes,
                   planes,
                   blocks,
                   spatial_stride=1,
                   temporal_stride=1,
                   dilation=1,
                   inflate_freq=1,
                   inflate_style='3x1x1',
                   nonlocal_freq=1,
                   nonlocal_cfg=None,
                   norm_layer=BatchNorm3d,
                   norm_kwargs=None,
                   layer_name=''):
    inflate_freq = inflate_freq if not isinstance(inflate_freq, int) else (inflate_freq,) * blocks
    nonlocal_freq = nonlocal_freq if not isinstance(nonlocal_freq, int) else (nonlocal_freq,) * blocks
    assert len(inflate_freq) == blocks
    assert len(nonlocal_freq) == blocks

    downsample = None
    if spatial_stride != 1 or inplanes != planes * block.expansion:
        downsample = nn.Sequential(
            nn.Conv3d(in_channels=inplanes,
                      out_channels=planes * block.expansion,
                      kernel_size=1,
                      stride=(temporal_stride, spatial_stride, spatial_stride),
                      bias=False),
            norm_layer(num_features=planes * block.expansion, **({} if norm_kwargs is None else norm_kwargs)))

    layers = []
    cnt = 0
    layers.append(block(inplanes=inplanes,
                        planes=planes,
                        spatial_stride=spatial_stride,
                        temporal_stride=temporal_stride,
                        dilation=dilation,
                        downsample=downsample,
                        if_inflate=(inflate_freq[0] == 1),
                        inflate_style=inflate_style,
                        if_nonlocal=(nonlocal_freq[0] == 1),
                        nonlocal_cfg=nonlocal_cfg,
                        layer_name='%d_' % cnt))

    cnt += 1
    inplanes = planes * block.expansion
    for i in range(1, blocks):
        layers.append(block(inplanes=inplanes,
                            planes=planes,
                            spatial_stride=1,
                            temporal_stride=1,
                            dilation=dilation,
                            if_inflate=(inflate_freq[i] == 1),
                            inflate_style=inflate_style,
                            if_nonlocal=(nonlocal_freq[i] == 1),
                            nonlocal_cfg=nonlocal_cfg,
                            layer_name='%d_' % cnt))
        cnt += 1
    return nn.Sequential(*layers)


class I3D_ResNetV1(nn.Module):
    """ResNet_I3D backbone.
    Inflated 3D model (I3D) from
    `"Quo Vadis, Action Recognition? A New Model and the Kinetics Dataset"
    <https://arxiv.org/abs/1705.07750>`_ paper.
    Args:
        depth (int): Depth of ResNet, from {18, 34, 50, 101, 152}.
        num_stages (int): ResNet stages, normally 4.
        strides (Sequence[int]): Strides of the first block of each stage.
        dilations (Sequence[int]): Dilation of each stage.
        out_indices (Sequence[int]): Output from which stages.
        frozen_stages (int): Stages to be frozen (all param fixed). -1 means
            not freezing any parameters.
        bn_eval (bool): Whether to set BN layers to eval mode, namely, freeze
            running stats (mean and var).
        bn_frozen (bool): Whether to freeze weight and bias of BN layers.
    """

    arch_settings = {
        18: (BasicBlock, (2, 2, 2, 2)),
        34: (BasicBlock, (3, 4, 6, 3)),
        50: (Bottleneck, (3, 4, 6, 3)),
        101: (Bottleneck, (3, 4, 23, 3)),
        152: (Bottleneck, (3, 8, 36, 3))
    }

    def __init__(self,
                 num_classes,
                 depth,
                 num_stages=4,
                 pretrained=False,
                 pretrained_base=True,
                 feat_ext=False,
                 num_segment=1,
                 num_crop=1,
                 spatial_strides=(1, 2, 2, 2),
                 temporal_strides=(1, 1, 1, 1),
                 dilations=(1, 1, 1, 1),
                 out_indices=(0, 1, 2, 3),
                 conv1_kernel_t=5,
                 conv1_stride_t=2,
                 pool1_kernel_t=1,
                 pool1_stride_t=2,
                 inflate_freq=(1, 1, 1, 1),
                 inflate_stride=(1, 1, 1, 1),
                 inflate_style='3x1x1',
                 nonlocal_stages=(-1,),
                 nonlocal_freq=(0, 1, 1, 0),
                 nonlocal_cfg=None,
                 bn_eval=True,
                 bn_frozen=False,
                 partial_bn=False,
                 frozen_stages=-1,
                 dropout_ratio=0.5,
                 init_std=0.01,
                 norm_layer=BatchNorm3d,
                 norm_kwargs=None,
                 ctx=None,
                 **kwargs):
        super(I3D_ResNetV1, self).__init__()

        if depth not in self.arch_settings:
            raise KeyError('invalid depth {} for resnet'.format(depth))

        self.num_classes = num_classes
        self.depth = depth
        self.num_stages = num_stages
        self.pretrained = pretrained
        self.pretrained_base = pretrained_base
        self.feat_ext = feat_ext
        self.num_segment = num_segment
        self.num_crop = num_crop
        self.spatial_strides = spatial_strides
        self.temporal_strides = temporal_strides
        self.dilations = dilations
        assert len(spatial_strides) == len(temporal_strides) == len(dilations) == num_stages
        self.out_indices = out_indices
        assert max(out_indices) < num_stages
        self.inflate_freqs = inflate_freq if not isinstance(inflate_freq, int) else (inflate_freq,) * num_stages
        self.inflate_style = inflate_style
        self.nonlocal_stages = nonlocal_stages
        self.nonlocal_freqs = nonlocal_freq if not isinstance(nonlocal_freq, int) else (nonlocal_freq,) * num_stages
        self.nonlocal_cfg = nonlocal_cfg
        self.bn_eval = bn_eval
        self.bn_frozen = bn_frozen
        self.partial_bn = partial_bn
        self.frozen_stages = frozen_stages
        self.dropout_ratio = dropout_ratio
        self.init_std = init_std

        self.block, stage_blocks = self.arch_settings[depth]
        self.stage_blocks = stage_blocks[:num_stages]
        self.inplanes = 64

        if self.bn_frozen:
            if norm_kwargs is not None:
                norm_kwargs['use_global_stats'] = True
            else:
                norm_kwargs = {}
                norm_kwargs['use_global_stats'] = True

        self.first_stage = nn.Sequential(
            nn.Conv3d(in_channels=3, out_channels=64, kernel_size=(conv1_kernel_t, 7, 7),
                      stride=(conv1_stride_t, 2, 2), padding=((conv1_kernel_t - 1) // 2, 3, 3),
                      bias=False),
            norm_layer(num_features=64, **({} if norm_kwargs is None else norm_kwargs)),
            nn.ReLU(inplace=False),
            nn.MaxPool3d(kernel_size=(pool1_kernel_t, 3, 3), stride=(pool1_stride_t, 2, 2),
                         padding=(pool1_kernel_t // 2, 1, 1)))

        self.pool2 = nn.MaxPool3d(kernel_size=(2, 1, 1), stride=(2, 1, 1), padding=(0, 0, 0))

        if self.partial_bn:
            if norm_kwargs is not None:
                norm_kwargs['use_global_stats'] = True
            else:
                norm_kwargs = {}
                norm_kwargs['use_global_stats'] = True

        res_layers = []
        for i, num_blocks in enumerate(self.stage_blocks):
            spatial_stride = spatial_strides[i]
            temporal_stride = temporal_strides[i]
            dilation = dilations[i]
            planes = 64 * 2 ** i
            layer_name = 'layer{}_'.format(i + 1)

            res_layer = make_res_layer(self.block,
                                       self.inplanes,
                                       planes,
                                       num_blocks,
                                       spatial_stride=spatial_stride,
                                       temporal_stride=temporal_stride,
                                       dilation=dilation,
                                       inflate_freq=self.inflate_freqs[i],
                                       inflate_style=self.inflate_style,
                                       nonlocal_freq=self.nonlocal_freqs[i],
                                       nonlocal_cfg=self.nonlocal_cfg if i in self.nonlocal_stages else None,
                                       norm_layer=norm_layer,
                                       norm_kwargs=norm_kwargs,
                                       layer_name=layer_name)
            self.inplanes = planes * self.block.expansion
            res_layers.append(res_layer)

        self.res_layers = nn.Sequential(*res_layers)
        self.feat_dim = self.block.expansion * 64 * 2 ** (len(self.stage_blocks) - 1)

        self.st_avg = nn.AdaptiveAvgPool3d(output_size=1)
        self.dp = nn.Dropout(self.dropout_ratio)
        self.fc = nn.Linear(in_features=self.feat_dim, out_features=num_classes)
        nn.init.normal_(self.fc.weight, 0, self.init_std)
        nn.init.constant_(self.fc.bias, 0)
        self.head = nn.Sequential(self.dp, self.fc)

        self.inflate_weights()

    def inflate_weights(self):
        """Inflate I3D network with its 2D ImageNet pretrained weights."""

        if not self.pretrained_base:
            raise RuntimeError("I3D models need to be inflated. Please set PRETRAINED_BASE to True in config.")

        if self.pretrained_base and not self.pretrained:
            import torchvision
            if self.depth == 50:
                R2D = torchvision.models.resnet50(pretrained=True, progress=True)
            elif self.depth == 101:
                R2D = torchvision.models.resnet101(pretrained=True, progress=True)
            else:
                raise RuntimeError("We only support ResNet50 and ResNet101 for I3D models at this moment.")

            # copy conv1
            conv1 = self.first_stage._modules['0']
            conv1_bn = self.first_stage._modules['1']

            conv1.weight.data.copy_(torch.unsqueeze(R2D.conv1.weight.data, dim=2).repeat(1, 1, 5, 1, 1))
            conv1_bn.weight.data.copy_(R2D.bn1.weight.data)
            conv1_bn.bias.data.copy_(R2D.bn1.bias.data)
            conv1_bn.running_mean.data.copy_(R2D.bn1.running_mean.data)
            conv1_bn.running_var.data.copy_(R2D.bn1.running_var.data)

            res2 = self.res_layers._modules['0']
            res3 = self.res_layers._modules['1']
            res4 = self.res_layers._modules['2']
            res5 = self.res_layers._modules['3']

            stages = [res2, res3, res4, res5]

            R2Dlayers = [R2D.layer1, R2D.layer2, R2D.layer3, R2D.layer4]

            for s, _ in enumerate(stages):
                res = stages[s]._modules
                count = 0

                for k, block in res.items():
                    if block.conv1.weight.data.shape[2] > 1:
                        block.conv1.weight.data.copy_(
                            torch.unsqueeze(R2Dlayers[s]._modules[str(k)].conv1.weight.data, dim=2).repeat(1, 1, 3, 1, 1))
                    else:
                        block.conv1.weight.data.copy_(
                            torch.unsqueeze(R2Dlayers[s]._modules[str(k)].conv1.weight.data, dim=2))
                    block.conv2.weight.data.copy_(torch.unsqueeze(R2Dlayers[s]._modules[str(k)].conv2.weight.data, dim=2))
                    block.conv3.weight.data.copy_(torch.unsqueeze(R2Dlayers[s]._modules[str(k)].conv3.weight.data, dim=2))

                    block.bn1.weight.data.copy_(R2Dlayers[s]._modules[str(k)].bn1.weight.data)
                    block.bn1.bias.data.copy_(R2Dlayers[s]._modules[str(k)].bn1.bias.data)
                    block.bn1.running_mean.data.copy_(R2Dlayers[s]._modules[str(k)].bn1.running_mean.data)
                    block.bn1.running_var.data.copy_(R2Dlayers[s]._modules[str(k)].bn1.running_var.data)

                    block.bn2.weight.data.copy_(R2Dlayers[s]._modules[str(k)].bn2.weight.data)
                    block.bn2.bias.data.copy_(R2Dlayers[s]._modules[str(k)].bn2.bias.data)
                    block.bn2.running_mean.data.copy_(R2Dlayers[s]._modules[str(k)].bn2.running_mean.data)
                    block.bn2.running_var.data.copy_(R2Dlayers[s]._modules[str(k)].bn2.running_var.data)

                    block.bn3.weight.data.copy_(R2Dlayers[s]._modules[str(k)].bn3.weight.data)
                    block.bn3.bias.data.copy_(R2Dlayers[s]._modules[str(k)].bn3.bias.data)
                    block.bn3.running_mean.data.copy_(R2Dlayers[s]._modules[str(k)].bn3.running_mean.data)
                    block.bn3.running_var.data.copy_(R2Dlayers[s]._modules[str(k)].bn3.running_var.data)

                    if block.downsample is not None:
                        down_conv = block.downsample._modules["0"]
                        down_bn = block.downsample._modules["1"]

                        down_conv.weight.data.copy_(
                            torch.unsqueeze(R2Dlayers[s]._modules[str(k)].downsample._modules['0'].weight.data, dim=2))
                        down_bn.weight.data.copy_(R2Dlayers[s]._modules[str(k)].downsample._modules['1'].weight.data)
                        down_bn.bias.data.copy_(R2Dlayers[s]._modules[str(k)].downsample._modules['1'].bias.data)
                        down_bn.running_mean.data.copy_(
                            R2Dlayers[s]._modules[str(k)].downsample._modules['1'].running_mean.data)
                        down_bn.running_var.data.copy_(
                            R2Dlayers[s]._modules[str(k)].downsample._modules['1'].running_var.data)
                    count += 1
            print("I3D weights inflated from pretrained C2D.")

    def forward(self, x):
        bs, _, _, _, _ = x.shape
        x = self.first_stage(x)

        for i, res_layer in enumerate(self.res_layers):
            x = res_layer(x)
            if i == 0:
                x = self.pool2(x)

        # spatial temporal average
        pooled_feat = self.st_avg(x)
        x = pooled_feat.view(bs, -1)

        if self.feat_ext:
            return x

        x = self.head(x)
        return x


def i3d_resnet50_v1_kinetics400(cfg):
    model = I3D_ResNetV1(num_classes=cfg.CONFIG.DATA.NUM_CLASSES,
                         depth=50,
                         pretrained=cfg.CONFIG.MODEL.PRETRAINED,
                         pretrained_base=cfg.CONFIG.MODEL.PRETRAINED_BASE,
                         feat_ext=cfg.CONFIG.INFERENCE.FEAT,
                         num_segment=cfg.CONFIG.DATA.NUM_SEGMENT,
                         num_crop=cfg.CONFIG.DATA.NUM_CROP,
                         out_indices=[3],
                         inflate_freq=((1, 1, 1), (1, 0, 1, 0), (1, 0, 1, 0, 1, 0), (0, 1, 0)),
                         bn_eval=cfg.CONFIG.MODEL.BN_EVAL,
                         partial_bn=cfg.CONFIG.MODEL.PARTIAL_BN,
                         bn_frozen=cfg.CONFIG.MODEL.BN_FROZEN)

    if cfg.CONFIG.MODEL.PRETRAINED:
        from ..model_store import get_model_file
        model.load_state_dict(torch.load(get_model_file('i3d_resnet50_v1_kinetics400',
                                                        tag=cfg.CONFIG.MODEL.PRETRAINED)))
    return model


def i3d_resnet101_v1_kinetics400(cfg):
    model = I3D_ResNetV1(num_classes=cfg.CONFIG.DATA.NUM_CLASSES,
                         depth=101,
                         pretrained=cfg.CONFIG.MODEL.PRETRAINED,
                         pretrained_base=cfg.CONFIG.MODEL.PRETRAINED_BASE,
                         feat_ext=cfg.CONFIG.INFERENCE.FEAT,
                         num_segment=cfg.CONFIG.DATA.NUM_SEGMENT,
                         num_crop=cfg.CONFIG.DATA.NUM_CROP,
                         out_indices=[3],
                         inflate_freq=((1, 1, 1),
                                       (1, 0, 1, 0),
                                       (1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1),
                                       (0, 1, 0)),
                         bn_eval=cfg.CONFIG.MODEL.BN_EVAL,
                         partial_bn=cfg.CONFIG.MODEL.PARTIAL_BN,
                         bn_frozen=cfg.CONFIG.MODEL.BN_FROZEN)

    if cfg.CONFIG.MODEL.PRETRAINED:
        from ..model_store import get_model_file
        model.load_state_dict(torch.load(get_model_file('i3d_resnet101_v1_kinetics400',
                                                        tag=cfg.CONFIG.MODEL.PRETRAINED)))
    return model


def i3d_nl5_resnet50_v1_kinetics400(cfg):
    model = I3D_ResNetV1(num_classes=cfg.CONFIG.DATA.NUM_CLASSES,
                         depth=50,
                         pretrained=cfg.CONFIG.MODEL.PRETRAINED,
                         pretrained_base=cfg.CONFIG.MODEL.PRETRAINED_BASE,
                         feat_ext=cfg.CONFIG.INFERENCE.FEAT,
                         num_segment=cfg.CONFIG.DATA.NUM_SEGMENT,
                         num_crop=cfg.CONFIG.DATA.NUM_CROP,
                         out_indices=[3],
                         inflate_freq=((1, 1, 1), (1, 0, 1, 0), (1, 0, 1, 0, 1, 0), (0, 1, 0)),
                         bn_eval=cfg.CONFIG.MODEL.BN_EVAL,
                         partial_bn=cfg.CONFIG.MODEL.PARTIAL_BN,
                         bn_frozen=cfg.CONFIG.MODEL.BN_FROZEN,
                         nonlocal_stages=(1, 2),
                         nonlocal_cfg=dict(nonlocal_type="gaussian"),
                         nonlocal_freq=((0, 0, 0), (0, 1, 0, 1), (0, 1, 0, 1, 0, 1), (0, 0, 0)))

    if cfg.CONFIG.MODEL.PRETRAINED:
        from ..model_store import get_model_file
        model.load_state_dict(torch.load(get_model_file('i3d_nl5_resnet50_v1_kinetics400',
                                                        tag=cfg.CONFIG.MODEL.PRETRAINED)))
    return model


def i3d_nl10_resnet50_v1_kinetics400(cfg):
    model = I3D_ResNetV1(num_classes=cfg.CONFIG.DATA.NUM_CLASSES,
                         depth=50,
                         pretrained=cfg.CONFIG.MODEL.PRETRAINED,
                         pretrained_base=cfg.CONFIG.MODEL.PRETRAINED_BASE,
                         feat_ext=cfg.CONFIG.INFERENCE.FEAT,
                         num_segment=cfg.CONFIG.DATA.NUM_SEGMENT,
                         num_crop=cfg.CONFIG.DATA.NUM_CROP,
                         out_indices=[3],
                         inflate_freq=((1, 1, 1), (1, 0, 1, 0), (1, 0, 1, 0, 1, 0), (0, 1, 0)),
                         bn_eval=cfg.CONFIG.MODEL.BN_EVAL,
                         partial_bn=cfg.CONFIG.MODEL.PARTIAL_BN,
                         bn_frozen=cfg.CONFIG.MODEL.BN_FROZEN,
                         nonlocal_stages=(1, 2),
                         nonlocal_cfg=dict(nonlocal_type="gaussian"),
                         nonlocal_freq=((0, 0, 0), (1, 1, 1, 1), (1, 1, 1, 1, 1, 1), (0, 0, 0)))

    if cfg.CONFIG.MODEL.PRETRAINED:
        from ..model_store import get_model_file
        model.load_state_dict(torch.load(get_model_file('i3d_nl10_resnet50_v1_kinetics400',
                                                        tag=cfg.CONFIG.MODEL.PRETRAINED)))
    return model


def i3d_nl5_resnet101_v1_kinetics400(cfg):
    model = I3D_ResNetV1(num_classes=cfg.CONFIG.DATA.NUM_CLASSES,
                         depth=101,
                         pretrained=cfg.CONFIG.MODEL.PRETRAINED,
                         pretrained_base=cfg.CONFIG.MODEL.PRETRAINED_BASE,
                         feat_ext=cfg.CONFIG.INFERENCE.FEAT,
                         num_segment=cfg.CONFIG.DATA.NUM_SEGMENT,
                         num_crop=cfg.CONFIG.DATA.NUM_CROP,
                         out_indices=[3],
                         inflate_freq=((1, 1, 1),
                                       (1, 0, 1, 0),
                                       (1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1),
                                       (0, 1, 0)),
                         bn_eval=cfg.CONFIG.MODEL.BN_EVAL,
                         partial_bn=cfg.CONFIG.MODEL.PARTIAL_BN,
                         bn_frozen=cfg.CONFIG.MODEL.BN_FROZEN,
                         nonlocal_stages=(1, 2),
                         nonlocal_cfg=dict(nonlocal_type="gaussian"),
                         nonlocal_freq=((0, 0, 0),
                                        (0, 1, 0, 1),
                                        (0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0),
                                        (0, 0, 0)))

    if cfg.CONFIG.MODEL.PRETRAINED:
        from ..model_store import get_model_file
        model.load_state_dict(torch.load(get_model_file('i3d_nl5_resnet101_v1_kinetics400',
                                                        tag=cfg.CONFIG.MODEL.PRETRAINED)))
    return model


def i3d_nl10_resnet101_v1_kinetics400(cfg):
    model = I3D_ResNetV1(num_classes=cfg.CONFIG.DATA.NUM_CLASSES,
                         depth=101,
                         pretrained=cfg.CONFIG.MODEL.PRETRAINED,
                         pretrained_base=cfg.CONFIG.MODEL.PRETRAINED_BASE,
                         feat_ext=cfg.CONFIG.INFERENCE.FEAT,
                         num_segment=cfg.CONFIG.DATA.NUM_SEGMENT,
                         num_crop=cfg.CONFIG.DATA.NUM_CROP,
                         out_indices=[3],
                         inflate_freq=((1, 1, 1),
                                       (1, 0, 1, 0),
                                       (1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1),
                                       (0, 1, 0)),
                         bn_eval=cfg.CONFIG.MODEL.BN_EVAL,
                         partial_bn=cfg.CONFIG.MODEL.PARTIAL_BN,
                         bn_frozen=cfg.CONFIG.MODEL.BN_FROZEN,
                         nonlocal_stages=(1, 2),
                         nonlocal_cfg=dict(nonlocal_type="gaussian"),
                         nonlocal_freq=((0, 0, 0),
                                        (1, 1, 1, 1),
                                        (0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1),
                                        (0, 0, 0)))
    if cfg.CONFIG.MODEL.PRETRAINED:
        from ..model_store import get_model_file
        model.load_state_dict(torch.load(get_model_file('i3d_nl10_resnet101_v1_kinetics400',
                                                        tag=cfg.CONFIG.MODEL.PRETRAINED)))
    return model


def i3d_resnet50_v1_sthsthv2(cfg):
    model = I3D_ResNetV1(num_classes=cfg.CONFIG.DATA.NUM_CLASSES,
                         depth=50,
                         pretrained=cfg.CONFIG.MODEL.PRETRAINED,
                         pretrained_base=cfg.CONFIG.MODEL.PRETRAINED_BASE,
                         feat_ext=cfg.CONFIG.INFERENCE.FEAT,
                         num_segment=cfg.CONFIG.DATA.NUM_SEGMENT,
                         num_crop=cfg.CONFIG.DATA.NUM_CROP,
                         out_indices=[3],
                         inflate_freq=((1, 1, 1), (1, 0, 1, 0), (1, 0, 1, 0, 1, 0), (0, 1, 0)),
                         bn_eval=cfg.CONFIG.MODEL.BN_EVAL,
                         partial_bn=cfg.CONFIG.MODEL.PARTIAL_BN,
                         bn_frozen=cfg.CONFIG.MODEL.BN_FROZEN)

    if cfg.CONFIG.MODEL.PRETRAINED:
        from ..model_store import get_model_file
        model.load_state_dict(torch.load(get_model_file('i3d_resnet50_v1_sthsthv2',
                                                        tag=cfg.CONFIG.MODEL.PRETRAINED)))
    return model


def i3d_resnet50_v1_custom(cfg):
    model = I3D_ResNetV1(num_classes=cfg.CONFIG.DATA.NUM_CLASSES,
                         depth=50,
                         pretrained=cfg.CONFIG.MODEL.PRETRAINED,
                         pretrained_base=cfg.CONFIG.MODEL.PRETRAINED_BASE,
                         feat_ext=cfg.CONFIG.INFERENCE.FEAT,
                         num_segment=cfg.CONFIG.DATA.NUM_SEGMENT,
                         num_crop=cfg.CONFIG.DATA.NUM_CROP,
                         out_indices=[3],
                         inflate_freq=((1, 1, 1), (1, 0, 1, 0), (1, 0, 1, 0, 1, 0), (0, 1, 0)),
                         bn_eval=cfg.CONFIG.MODEL.BN_EVAL,
                         partial_bn=cfg.CONFIG.MODEL.PARTIAL_BN,
                         bn_frozen=cfg.CONFIG.MODEL.BN_FROZEN)

    if cfg.CONFIG.MODEL.PRETRAINED:
        from ..model_store import get_model_file
        state_dict = torch.load(get_model_file('i3d_resnet50_v1_kinetics400', tag=cfg.CONFIG.MODEL.PRETRAINED))
        for k in list(state_dict.keys()):
            # retain only backbone up to before the classification layer
            if k.startswith('fc') or k.startswith('head'):
                del state_dict[k]

        msg = model.load_state_dict(state_dict, strict=False)
        assert set(msg.missing_keys) == {'fc.weight', 'fc.bias', 'head.1.weight', 'head.1.bias'}
        print("=> Initialized from an I3D model pretrained on Kinetcis400 dataset")
    return model



"""Pre init checks"""
import sys

def _require_mxnet_version(mx_version, max_mx_version='2.0.0'):
    try:
        import mxnet as mx
        from distutils.version import LooseVersion
        if LooseVersion(mx.__version__) < LooseVersion(mx_version) or \
            LooseVersion(mx.__version__) >= LooseVersion(max_mx_version):
            version_str = '>={},<{}'.format(mx_version, max_mx_version)
            msg = (
                "Legacy mxnet=={0} detected, some modules may not work properly. "
                "mxnet{1} is required. You can use pip to upgrade mxnet "
                "`pip install -U 'mxnet{1}'` "
                "or `pip install -U 'mxnet-cu100{1}'`\
                ").format(mx.__version__, version_str)
            raise RuntimeError(msg)
    except ImportError as ex:
        raise ImportError(
            "Unable to import dependency mxnet. "
            "A quick tip is to install via "
            "`pip install 'mxnet-cu100<{}'`. "
            "please refer to https://gluon-cv.mxnet.io/#installation for details.".format(
                max_mx_version)) from ex

def _require_pytorch_version(torch_version, max_torch_version='2.0.0'):
    try:
        import torch
        from distutils.version import LooseVersion
        if LooseVersion(torch.__version__) < LooseVersion(torch_version) or \
            LooseVersion(torch.__version__) >= LooseVersion(max_torch_version):
            version_str = '>={},<{}'.format(torch_version, max_torch_version)
            msg = (
                "Legacy torch=={0} detected, some modules may not work properly. "
                "torch{1} is required. You can use pip or conda to upgrade".format(torch.__version__, version_str))
            raise RuntimeError(msg)
    except ImportError as ex:
        raise ImportError(
            "Unable to import dependency pytorch. Please use pip or conda to install.") from ex

def _deprecate_python2():
    if sys.version_info[0] < 3:
        msg = 'Python2 has reached the end of its life on January 1st, 2020. ' + \
            'GluonCV has now dropped support for Python 2.'
        raise DeprecationWarning(msg)



"""FCOS object detector"""
# pylint: disable=all
import logging
import math
from typing import List, Dict
import torch
from torch import nn
from torch.nn import functional as F

from ...nn.shape_spec import ShapeSpec
from ...nn.deform_conv import DeformConvWithChangeableStride
from ...nn.group_norm import NaiveGroupNorm

from ...data.structures import Instances, Boxes
from ...nn.focal_loss import sigmoid_focal_loss_jit

from ...nn.nms import ml_nms
from ...utils.loss import IOULoss
from ...utils.comm import get_world_size, reduce_sum, compute_locations


__all__ = ["FCOS"]

INF = 100000000

def cat(tensors: List[torch.Tensor], dim: int = 0):
    """
    Efficient version of torch.cat that avoids a copy if there is only a single element in a list
    """
    assert isinstance(tensors, (list, tuple))
    if len(tensors) == 1:
        return tensors[0]
    return torch.cat(tensors, dim)


class Scale(nn.Module):
    def __init__(self, init_value=1.0):
        super(Scale, self).__init__()
        self.scale = nn.Parameter(torch.FloatTensor([init_value]))

    def forward(self, input):
        return input * self.scale


class ModuleListDial(nn.ModuleList):
    def __init__(self, modules=None):
        super(ModuleListDial, self).__init__(modules)
        self.cur_position = 0

    def forward(self, x):
        result = self[self.cur_position](x)
        self.cur_position += 1
        if self.cur_position >= len(self):
            self.cur_position = 0
        return result


@PROPOSAL_GENERATOR_REGISTRY.register()
class FCOS(nn.Module):
    """
    Implement FCOS (https://arxiv.org/abs/1904.01355).
    """
    def __init__(self, cfg, input_shape: Dict[str, ShapeSpec]):
        super().__init__()
        self.in_features = cfg.CONFIG.MODEL.FCOS.IN_FEATURES
        self.fpn_strides = cfg.CONFIG.MODEL.FCOS.FPN_STRIDES
        self.yield_proposal = cfg.CONFIG.MODEL.FCOS.YIELD_PROPOSAL

        self.fcos_head = FCOSHead(cfg, [input_shape[f] for f in self.in_features])
        self.in_channels_to_top_module = self.fcos_head.in_channels_to_top_module

        self.fcos_outputs = FCOSOutputs(cfg)

    def forward_head(self, features, top_module=None):
        features = [features[f] for f in self.in_features]
        pred_class_logits, pred_deltas, pred_centerness, top_feats, bbox_towers = self.fcos_head(
            features, top_module, self.yield_proposal)
        return pred_class_logits, pred_deltas, pred_centerness, top_feats, bbox_towers

    def forward(self, images, features, gt_instances=None, top_module=None):
        """
        Arguments:
            images (list[Tensor] or ImageList): images to be processed
            targets (list[BoxList]): ground-truth boxes present in the image (optional)

        Returns:
            result (list[BoxList] or dict[Tensor]): the output from the model.
                During training, it returns a dict[Tensor] which contains the losses.
                During testing, it returns list[BoxList] contains additional fields
                like `scores`, `labels` and `mask` (for Mask R-CNN models).

        """
        features = [features[f] for f in self.in_features]
        locations = self.compute_locations(features)
        logits_pred, reg_pred, ctrness_pred, top_feats, bbox_towers = self.fcos_head(
            features, top_module, self.yield_proposal
        )

        results = {}
        if self.yield_proposal:
            results["features"] = {
                f: b for f, b in zip(self.in_features, bbox_towers)
            }

        if self.training:
            results, losses = self.fcos_outputs.losses(
                logits_pred, reg_pred, ctrness_pred,
                locations, gt_instances, top_feats
            )

            if self.yield_proposal:
                with torch.no_grad():
                    results["proposals"] = self.fcos_outputs.predict_proposals(
                        logits_pred, reg_pred, ctrness_pred,
                        locations, images.image_sizes, top_feats
                    )
            return results, losses
        else:
            results = self.fcos_outputs.predict_proposals(
                logits_pred, reg_pred, ctrness_pred,
                locations, images.image_sizes, top_feats
            )

            return results, {}

    def compute_locations(self, features):
        locations = []
        for level, feature in enumerate(features):
            h, w = feature.size()[-2:]
            locations_per_level = compute_locations(
                h, w, self.fpn_strides[level],
                feature.device
            )
            locations.append(locations_per_level)
        return locations


class FCOSHead(nn.Module):
    def __init__(self, cfg, input_shape: List[ShapeSpec]):
        """
        Arguments:
            in_channels (int): number of channels of the input feature
        """
        super().__init__()
        # TODO: Implement the sigmoid version first.
        self.num_classes = cfg.CONFIG.MODEL.FCOS.NUM_CLASSES
        self.fpn_strides = cfg.CONFIG.MODEL.FCOS.FPN_STRIDES
        head_configs = {"cls": (cfg.CONFIG.MODEL.FCOS.NUM_CLS_CONVS,
                                cfg.CONFIG.MODEL.FCOS.USE_DEFORMABLE),
                        "bbox": (cfg.CONFIG.MODEL.FCOS.NUM_BOX_CONVS,
                                 cfg.CONFIG.MODEL.FCOS.USE_DEFORMABLE),
                        "share": (cfg.CONFIG.MODEL.FCOS.NUM_SHARE_CONVS,
                                  False)}
        norm = None if cfg.CONFIG.MODEL.FCOS.NORM == "none" else cfg.CONFIG.MODEL.FCOS.NORM
        self.num_levels = len(input_shape)

        in_channels = [s.channels for s in input_shape]
        assert len(set(in_channels)) == 1, "Each level must have the same channel!"
        in_channels = in_channels[0]

        self.in_channels_to_top_module = in_channels

        for head in head_configs:
            tower = []
            num_convs, use_deformable = head_configs[head]
            for i in range(num_convs):
                if use_deformable and i == num_convs - 1:
                    raise NotImplementedError("Deformable Conv v2 is not supported yet!")
                else:
                    conv_func = nn.Conv2d
                tower.append(conv_func(
                    in_channels, in_channels,
                    kernel_size=3, stride=1,
                    padding=1, bias=True
                ))
                if norm == "GN":
                    tower.append(nn.GroupNorm(32, in_channels))
                elif norm == "NaiveGN":
                    tower.append(NaiveGroupNorm(32, in_channels))
                elif norm == "BN":
                    tower.append(ModuleListDial([
                        nn.BatchNorm2d(in_channels) for _ in range(self.num_levels)
                    ]))
                elif norm == "SyncBN":
                    tower.append(ModuleListDial([
                        NaiveSyncBatchNorm(in_channels) for _ in range(self.num_levels)
                    ]))
                tower.append(nn.ReLU())
            self.add_module('{}_tower'.format(head),
                            nn.Sequential(*tower))

        self.cls_logits = nn.Conv2d(
            in_channels, self.num_classes,
            kernel_size=3, stride=1,
            padding=1
        )
        self.bbox_pred = nn.Conv2d(
            in_channels, 4, kernel_size=3,
            stride=1, padding=1
        )
        self.ctrness = nn.Conv2d(
            in_channels, 1, kernel_size=3,
            stride=1, padding=1
        )

        if cfg.CONFIG.MODEL.FCOS.USE_SCALE:
            self.scales = nn.ModuleList([Scale(init_value=1.0) for _ in range(self.num_levels)])
        else:
            self.scales = None

        for modules in [
            self.cls_tower, self.bbox_tower,
            self.share_tower, self.cls_logits,
            self.bbox_pred, self.ctrness
        ]:
            for l in modules.modules():
                if isinstance(l, nn.Conv2d):
                    torch.nn.init.normal_(l.weight, std=0.01)
                    torch.nn.init.constant_(l.bias, 0)

        # initialize the bias for focal loss
        prior_prob = cfg.CONFIG.MODEL.FCOS.PRIOR_PROB
        bias_value = -math.log((1 - prior_prob) / prior_prob)
        torch.nn.init.constant_(self.cls_logits.bias, bias_value)

    def forward(self, x, top_module=None, yield_bbox_towers=False):
        logits = []
        bbox_reg = []
        ctrness = []
        top_feats = []
        bbox_towers = []
        for l, feature in enumerate(x):
            feature = self.share_tower(feature)
            cls_tower = self.cls_tower(feature)
            bbox_tower = self.bbox_tower(feature)
            if yield_bbox_towers:
                bbox_towers.append(bbox_tower)

            logits.append(self.cls_logits(cls_tower))
            ctrness.append(self.ctrness(bbox_tower))
            reg = self.bbox_pred(bbox_tower)
            if self.scales is not None:
                reg = self.scales[l](reg)
            # Note that we use relu, as in the improved FCOS, instead of exp.
            bbox_reg.append(F.relu(reg))
            if top_module is not None:
                top_feats.append(top_module(bbox_tower))
        return logits, bbox_reg, ctrness, top_feats, bbox_towers

"""
Shape shorthand in this module:

    N: number of images in the minibatch
    L: number of feature maps per image on which RPN is run
    Hi, Wi: height and width of the i-th feature map
    4: size of the box parameterization

Naming convention:

    labels: refers to the ground-truth class of an position.

    reg_targets: refers to the 4-d (left, top, right, bottom) distances that parameterize the ground-truth box.

    logits_pred: predicted classification scores in [-inf, +inf];

    reg_pred: the predicted (left, top, right, bottom), corresponding to reg_targets

    ctrness_pred: predicted centerness scores

"""


def compute_ctrness_targets(reg_targets):
    if len(reg_targets) == 0:
        return reg_targets.new_zeros(len(reg_targets))
    left_right = reg_targets[:, [0, 2]]
    top_bottom = reg_targets[:, [1, 3]]
    ctrness = (left_right.min(dim=-1)[0] / left_right.max(dim=-1)[0]) * \
                 (top_bottom.min(dim=-1)[0] / top_bottom.max(dim=-1)[0])
    return torch.sqrt(ctrness)


class FCOSOutputs(nn.Module):
    def __init__(self, cfg):
        super(FCOSOutputs, self).__init__()

        self.focal_loss_alpha = cfg.CONFIG.MODEL.FCOS.LOSS_ALPHA
        self.focal_loss_gamma = cfg.CONFIG.MODEL.FCOS.LOSS_GAMMA
        self.center_sample = cfg.CONFIG.MODEL.FCOS.CENTER_SAMPLE
        self.radius = cfg.CONFIG.MODEL.FCOS.POS_RADIUS
        self.pre_nms_thresh_train = cfg.CONFIG.MODEL.FCOS.INFERENCE_TH_TRAIN
        self.pre_nms_topk_train = cfg.CONFIG.MODEL.FCOS.PRE_NMS_TOPK_TRAIN
        self.post_nms_topk_train = cfg.CONFIG.MODEL.FCOS.POST_NMS_TOPK_TRAIN
        self.loc_loss_func = IOULoss(cfg.CONFIG.MODEL.FCOS.LOC_LOSS_TYPE)

        self.pre_nms_thresh_test = cfg.CONFIG.MODEL.FCOS.INFERENCE_TH_TEST
        self.pre_nms_topk_test = cfg.CONFIG.MODEL.FCOS.PRE_NMS_TOPK_TEST
        self.post_nms_topk_test = cfg.CONFIG.MODEL.FCOS.POST_NMS_TOPK_TEST
        self.nms_thresh = cfg.CONFIG.MODEL.FCOS.NMS_TH
        self.thresh_with_ctr = cfg.CONFIG.MODEL.FCOS.THRESH_WITH_CTR

        self.num_classes = cfg.CONFIG.MODEL.FCOS.NUM_CLASSES
        self.strides = cfg.CONFIG.MODEL.FCOS.FPN_STRIDES

        # generate sizes of interest
        soi = []
        prev_size = -1
        for s in cfg.CONFIG.MODEL.FCOS.SIZES_OF_INTEREST:
            soi.append([prev_size, s])
            prev_size = s
        soi.append([prev_size, INF])
        self.sizes_of_interest = soi

    def _transpose(self, training_targets, num_loc_list):
        '''
        This function is used to transpose image first training targets to level first ones
        :return: level first training targets
        '''
        for im_i in range(len(training_targets)):
            training_targets[im_i] = torch.split(
                training_targets[im_i], num_loc_list, dim=0
            )

        targets_level_first = []
        for targets_per_level in zip(*training_targets):
            targets_level_first.append(
                torch.cat(targets_per_level, dim=0)
            )
        return targets_level_first

    def _get_ground_truth(self, locations, gt_instances):
        num_loc_list = [len(loc) for loc in locations]

        # compute locations to size ranges
        loc_to_size_range = []
        for l, loc_per_level in enumerate(locations):
            loc_to_size_range_per_level = loc_per_level.new_tensor(self.sizes_of_interest[l])
            loc_to_size_range.append(
                loc_to_size_range_per_level[None].expand(num_loc_list[l], -1)
            )

        loc_to_size_range = torch.cat(loc_to_size_range, dim=0)
        locations = torch.cat(locations, dim=0)

        training_targets = self.compute_targets_for_locations(
            locations, gt_instances, loc_to_size_range, num_loc_list
        )

        training_targets["locations"] = [locations.clone() for _ in range(len(gt_instances))]
        training_targets["im_inds"] = [
            locations.new_ones(locations.size(0), dtype=torch.long) * i for i in range(len(gt_instances))
        ]

        # transpose im first training_targets to level first ones
        training_targets = {
            k: self._transpose(v, num_loc_list) for k, v in training_targets.items()
        }

        training_targets["fpn_levels"] = [
            loc.new_ones(len(loc), dtype=torch.long) * level
            for level, loc in enumerate(training_targets["locations"])
        ]

        # we normalize reg_targets by FPN's strides here
        reg_targets = training_targets["reg_targets"]
        for l in range(len(reg_targets)):
            reg_targets[l] = reg_targets[l] / float(self.strides[l])

        return training_targets

    def get_sample_region(self, boxes, strides, num_loc_list, loc_xs, loc_ys, bitmasks=None, radius=1):
        if bitmasks is not None:
            _, h, w = bitmasks.size()

            ys = torch.arange(0, h, dtype=torch.float32, device=bitmasks.device)
            xs = torch.arange(0, w, dtype=torch.float32, device=bitmasks.device)

            m00 = bitmasks.sum(dim=-1).sum(dim=-1).clamp(min=1e-6)
            m10 = (bitmasks * xs).sum(dim=-1).sum(dim=-1)
            m01 = (bitmasks * ys[:, None]).sum(dim=-1).sum(dim=-1)
            center_x = m10 / m00
            center_y = m01 / m00
        else:
            center_x = boxes[..., [0, 2]].sum(dim=-1) * 0.5
            center_y = boxes[..., [1, 3]].sum(dim=-1) * 0.5

        num_gts = boxes.shape[0]
        K = len(loc_xs)
        boxes = boxes[None].expand(K, num_gts, 4)
        center_x = center_x[None].expand(K, num_gts)
        center_y = center_y[None].expand(K, num_gts)
        center_gt = boxes.new_zeros(boxes.shape)
        # no gt
        if center_x.numel() == 0 or center_x[..., 0].sum() == 0:
            return loc_xs.new_zeros(loc_xs.shape, dtype=torch.uint8)
        beg = 0
        for level, num_loc in enumerate(num_loc_list):
            end = beg + num_loc
            stride = strides[level] * radius
            xmin = center_x[beg:end] - stride
            ymin = center_y[beg:end] - stride
            xmax = center_x[beg:end] + stride
            ymax = center_y[beg:end] + stride
            # limit sample region in gt
            center_gt[beg:end, :, 0] = torch.where(xmin > boxes[beg:end, :, 0], xmin, boxes[beg:end, :, 0])
            center_gt[beg:end, :, 1] = torch.where(ymin > boxes[beg:end, :, 1], ymin, boxes[beg:end, :, 1])
            center_gt[beg:end, :, 2] = torch.where(xmax > boxes[beg:end, :, 2], boxes[beg:end, :, 2], xmax)
            center_gt[beg:end, :, 3] = torch.where(ymax > boxes[beg:end, :, 3], boxes[beg:end, :, 3], ymax)
            beg = end
        left = loc_xs[:, None] - center_gt[..., 0]
        right = center_gt[..., 2] - loc_xs[:, None]
        top = loc_ys[:, None] - center_gt[..., 1]
        bottom = center_gt[..., 3] - loc_ys[:, None]
        center_bbox = torch.stack((left, top, right, bottom), -1)
        inside_gt_bbox_mask = center_bbox.min(-1)[0] > 0
        return inside_gt_bbox_mask

    def compute_targets_for_locations(self, locations, targets, size_ranges, num_loc_list):
        labels = []
        reg_targets = []
        target_inds = []
        xs, ys = locations[:, 0], locations[:, 1]

        num_targets = 0
        for im_i in range(len(targets)):
            targets_per_im = targets[im_i]
            bboxes = targets_per_im.gt_boxes.tensor
            labels_per_im = targets_per_im.gt_classes

            # no gt
            if bboxes.numel() == 0:
                labels.append(labels_per_im.new_zeros(locations.size(0)) + self.num_classes)
                reg_targets.append(locations.new_zeros((locations.size(0), 4)))
                target_inds.append(labels_per_im.new_zeros(locations.size(0)) - 1)
                continue

            area = targets_per_im.gt_boxes.area()

            l = xs[:, None] - bboxes[:, 0][None]
            t = ys[:, None] - bboxes[:, 1][None]
            r = bboxes[:, 2][None] - xs[:, None]
            b = bboxes[:, 3][None] - ys[:, None]
            reg_targets_per_im = torch.stack([l, t, r, b], dim=2)

            if self.center_sample:
                if targets_per_im.has("gt_bitmasks_full"):
                    bitmasks = targets_per_im.gt_bitmasks_full
                else:
                    bitmasks = None
                is_in_boxes = self.get_sample_region(
                    bboxes, self.strides, num_loc_list, xs, ys,
                    bitmasks=bitmasks, radius=self.radius
                )
            else:
                is_in_boxes = reg_targets_per_im.min(dim=2)[0] > 0

            max_reg_targets_per_im = reg_targets_per_im.max(dim=2)[0]
            # limit the regression range for each location
            is_cared_in_the_level = \
                (max_reg_targets_per_im >= size_ranges[:, [0]]) & \
                (max_reg_targets_per_im <= size_ranges[:, [1]])

            locations_to_gt_area = area[None].repeat(len(locations), 1)
            locations_to_gt_area[is_in_boxes == 0] = INF
            locations_to_gt_area[is_cared_in_the_level == 0] = INF

            # if there are still more than one objects for a location,
            # we choose the one with minimal area
            locations_to_min_area, locations_to_gt_inds = locations_to_gt_area.min(dim=1)

            reg_targets_per_im = reg_targets_per_im[range(len(locations)), locations_to_gt_inds]
            target_inds_per_im = locations_to_gt_inds + num_targets
            num_targets += len(targets_per_im)

            labels_per_im = labels_per_im[locations_to_gt_inds]
            labels_per_im[locations_to_min_area == INF] = self.num_classes

            labels.append(labels_per_im)
            reg_targets.append(reg_targets_per_im)
            target_inds.append(target_inds_per_im)

        return {
            "labels": labels,
            "reg_targets": reg_targets,
            "target_inds": target_inds
        }

    def losses(self, logits_pred, reg_pred, ctrness_pred, locations, gt_instances, top_feats=None):
        """
        Return the losses from a set of FCOS predictions and their associated ground-truth.

        Returns:
            dict[loss name -> loss value]: A dict mapping from loss name to loss value.
        """

        training_targets = self._get_ground_truth(locations, gt_instances)

        # Collect all logits and regression predictions over feature maps
        # and images to arrive at the same shape as the labels and targets
        # The final ordering is L, N, H, W from slowest to fastest axis.

        instances = Instances((0, 0))
        instances.labels = cat([
            # Reshape: (N, 1, Hi, Wi) -> (N*Hi*Wi,)
            x.reshape(-1) for x in training_targets["labels"]
        ], dim=0)
        instances.gt_inds = cat([
            # Reshape: (N, 1, Hi, Wi) -> (N*Hi*Wi,)
            x.reshape(-1) for x in training_targets["target_inds"]
        ], dim=0)
        instances.im_inds = cat([
            x.reshape(-1) for x in training_targets["im_inds"]
        ], dim=0)
        instances.reg_targets = cat([
            # Reshape: (N, Hi, Wi, 4) -> (N*Hi*Wi, 4)
            x.reshape(-1, 4) for x in training_targets["reg_targets"]
        ], dim=0,)
        instances.locations = cat([
            x.reshape(-1, 2) for x in training_targets["locations"]
        ], dim=0)
        instances.fpn_levels = cat([
            x.reshape(-1) for x in training_targets["fpn_levels"]
        ], dim=0)

        instances.logits_pred = cat([
            # Reshape: (N, C, Hi, Wi) -> (N, Hi, Wi, C) -> (N*Hi*Wi, C)
            x.permute(0, 2, 3, 1).reshape(-1, self.num_classes) for x in logits_pred
        ], dim=0,)
        instances.reg_pred = cat([
            # Reshape: (N, B, Hi, Wi) -> (N, Hi, Wi, B) -> (N*Hi*Wi, B)
            x.permute(0, 2, 3, 1).reshape(-1, 4) for x in reg_pred
        ], dim=0,)
        instances.ctrness_pred = cat([
            # Reshape: (N, 1, Hi, Wi) -> (N*Hi*Wi,)
            x.permute(0, 2, 3, 1).reshape(-1) for x in ctrness_pred
        ], dim=0,)

        if len(top_feats) > 0:
            instances.top_feats = cat([
                # Reshape: (N, -1, Hi, Wi) -> (N*Hi*Wi, -1)
                x.permute(0, 2, 3, 1).reshape(-1, x.size(1)) for x in top_feats
            ], dim=0,)

        return self.fcos_losses(instances)

    def fcos_losses(self, instances):
        num_classes = instances.logits_pred.size(1)
        assert num_classes == self.num_classes

        labels = instances.labels.flatten()

        pos_inds = torch.nonzero(labels != num_classes).squeeze(1)
        num_pos_local = pos_inds.numel()
        num_gpus = get_world_size()
        total_num_pos = reduce_sum(pos_inds.new_tensor([num_pos_local])).item()
        num_pos_avg = max(total_num_pos / num_gpus, 1.0)

        # prepare one_hot
        class_target = torch.zeros_like(instances.logits_pred)
        class_target[pos_inds, labels[pos_inds]] = 1

        class_loss = sigmoid_focal_loss_jit(
            instances.logits_pred,
            class_target,
            alpha=self.focal_loss_alpha,
            gamma=self.focal_loss_gamma,
            reduction="sum",
        ) / num_pos_avg

        instances = instances[pos_inds]
        instances.pos_inds = pos_inds

        ctrness_targets = compute_ctrness_targets(instances.reg_targets)
        ctrness_targets_sum = ctrness_targets.sum()
        loss_denorm = max(reduce_sum(ctrness_targets_sum).item() / num_gpus, 1e-6)
        instances.gt_ctrs = ctrness_targets

        if pos_inds.numel() > 0:
            reg_loss = self.loc_loss_func(
                instances.reg_pred,
                instances.reg_targets,
                ctrness_targets
            ) / loss_denorm

            ctrness_loss = F.binary_cross_entropy_with_logits(
                instances.ctrness_pred,
                ctrness_targets,
                reduction="sum"
            ) / num_pos_avg
        else:
            reg_loss = instances.reg_pred.sum() * 0
            ctrness_loss = instances.ctrness_pred.sum() * 0

        losses = {
            "loss_fcos_cls": class_loss,
            "loss_fcos_loc": reg_loss,
            "loss_fcos_ctr": ctrness_loss
        }
        extras = {
            "instances": instances,
            "loss_denorm": loss_denorm
        }
        return extras, losses

    def predict_proposals(
            self, logits_pred, reg_pred, ctrness_pred,
            locations, image_sizes, top_feats=None
    ):
        if self.training:
            self.pre_nms_thresh = self.pre_nms_thresh_train
            self.pre_nms_topk = self.pre_nms_topk_train
            self.post_nms_topk = self.post_nms_topk_train
        else:
            self.pre_nms_thresh = self.pre_nms_thresh_test
            self.pre_nms_topk = self.pre_nms_topk_test
            self.post_nms_topk = self.post_nms_topk_test

        sampled_boxes = []

        bundle = {
            "l": locations, "o": logits_pred,
            "r": reg_pred, "c": ctrness_pred,
            "s": self.strides,
        }

        if len(top_feats) > 0:
            bundle["t"] = top_feats

        for i, per_bundle in enumerate(zip(*bundle.values())):
            # get per-level bundle
            per_bundle = dict(zip(bundle.keys(), per_bundle))
            # recall that during training, we normalize regression targets with FPN's stride.
            # we denormalize them here.
            l = per_bundle["l"]
            o = per_bundle["o"]
            r = per_bundle["r"] * per_bundle["s"]
            c = per_bundle["c"]
            t = per_bundle["t"] if "t" in bundle else None

            sampled_boxes.append(
                self.forward_for_single_feature_map(
                    l, o, r, c, image_sizes, t
                )
            )

            for per_im_sampled_boxes in sampled_boxes[-1]:
                per_im_sampled_boxes.fpn_levels = l.new_ones(
                    len(per_im_sampled_boxes), dtype=torch.long
                ) * i

        boxlists = list(zip(*sampled_boxes))
        boxlists = [Instances.cat(boxlist) for boxlist in boxlists]
        boxlists = self.select_over_all_levels(boxlists)

        return boxlists

    def forward_for_single_feature_map(
            self, locations, logits_pred, reg_pred,
            ctrness_pred, image_sizes, top_feat=None
    ):
        N, C, H, W = logits_pred.shape

        # put in the same format as locations
        logits_pred = logits_pred.view(N, C, H, W).permute(0, 2, 3, 1)
        logits_pred = logits_pred.reshape(N, -1, C).sigmoid()
        box_regression = reg_pred.view(N, 4, H, W).permute(0, 2, 3, 1)
        box_regression = box_regression.reshape(N, -1, 4)
        ctrness_pred = ctrness_pred.view(N, 1, H, W).permute(0, 2, 3, 1)
        ctrness_pred = ctrness_pred.reshape(N, -1).sigmoid()
        if top_feat is not None:
            top_feat = top_feat.view(N, -1, H, W).permute(0, 2, 3, 1)
            top_feat = top_feat.reshape(N, H * W, -1)

        # if self.thresh_with_ctr is True, we multiply the classification
        # scores with centerness scores before applying the threshold.
        if self.thresh_with_ctr:
            logits_pred = logits_pred * ctrness_pred[:, :, None]
        candidate_inds = logits_pred > self.pre_nms_thresh
        pre_nms_top_n = candidate_inds.view(N, -1).sum(1)
        pre_nms_top_n = pre_nms_top_n.clamp(max=self.pre_nms_topk)

        if not self.thresh_with_ctr:
            logits_pred = logits_pred * ctrness_pred[:, :, None]

        results = []
        for i in range(N):
            per_box_cls = logits_pred[i]
            per_candidate_inds = candidate_inds[i]
            per_box_cls = per_box_cls[per_candidate_inds]

            per_candidate_nonzeros = per_candidate_inds.nonzero()
            per_box_loc = per_candidate_nonzeros[:, 0]
            per_class = per_candidate_nonzeros[:, 1]

            per_box_regression = box_regression[i]
            per_box_regression = per_box_regression[per_box_loc]
            per_locations = locations[per_box_loc]
            if top_feat is not None:
                per_top_feat = top_feat[i]
                per_top_feat = per_top_feat[per_box_loc]

            per_pre_nms_top_n = pre_nms_top_n[i]

            if per_candidate_inds.sum().item() > per_pre_nms_top_n.item():
                per_box_cls, top_k_indices = \
                    per_box_cls.topk(per_pre_nms_top_n, sorted=False)
                per_class = per_class[top_k_indices]
                per_box_regression = per_box_regression[top_k_indices]
                per_locations = per_locations[top_k_indices]
                if top_feat is not None:
                    per_top_feat = per_top_feat[top_k_indices]

            detections = torch.stack([
                per_locations[:, 0] - per_box_regression[:, 0],
                per_locations[:, 1] - per_box_regression[:, 1],
                per_locations[:, 0] + per_box_regression[:, 2],
                per_locations[:, 1] + per_box_regression[:, 3],
            ], dim=1)

            boxlist = Instances(image_sizes[i])
            boxlist.pred_boxes = Boxes(detections)
            boxlist.scores = torch.sqrt(per_box_cls)
            boxlist.pred_classes = per_class
            boxlist.locations = per_locations
            if top_feat is not None:
                boxlist.top_feat = per_top_feat
            results.append(boxlist)

        return results

    def select_over_all_levels(self, boxlists):
        num_images = len(boxlists)
        results = []
        for i in range(num_images):
            # multiclass nms
            result = ml_nms(boxlists[i], self.nms_thresh)
            number_of_detections = len(result)

            # Limit to max_per_image detections **over all classes**
            if number_of_detections > self.post_nms_topk > 0:
                cls_scores = result.scores
                image_thresh, _ = torch.kthvalue(
                    cls_scores.cpu(),
                    number_of_detections - self.post_nms_topk + 1
                )
                keep = cls_scores >= image_thresh.item()
                keep = torch.nonzero(keep).squeeze(1)
                result = result[keep]
            results.append(result)
        return results
